import enum
from copy import deepcopy
from datetime import datetime, timezone

from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy.ext.declarative import declared_attr

from . import db
from .geo_data import (
    BaseData,
    DataAttribute,
    Feature,
    GeoData,
    GlobalData,
    StreamGeoData,
    WMSGeoData,
)
from .mixins import BaseModelMixin, TimestampMixin


class DataType(enum.Enum):
    geo_data = 1
    generator = 2
    global_data = 3
    data_stream = 4


input_assoc = db.Table(
    "data_input_association",
    db.Model.metadata,
    db.Column(
        "project_data_id",
        db.Integer,
        db.ForeignKey("project_data.id"),
        primary_key=True,
    ),
    db.Column(
        "input_data_id",
        db.Integer,
        db.ForeignKey("project_data.id"),
        primary_key=True,
    ),
)


class ProjectData(TimestampMixin, db.Model):
    __tablename__ = "project_data"

    data_type = db.Column(db.Enum(DataType))
    id = db.Column(db.Integer, primary_key=True)
    data_id = db.Column(db.Integer, db.ForeignKey("base_data.id"))
    last_update = db.Column(db.DateTime)
    project_id = db.Column(
        db.Integer, db.ForeignKey("project.id"), nullable=False
    )
    description = db.Column(db.String)
    name = db.Column(db.String)

    input_data = db.relationship(
        "ProjectData",
        secondary=input_assoc,
        primaryjoin=(id == input_assoc.c.project_data_id),
        secondaryjoin=(id == input_assoc.c.input_data_id),
        backref="output_data",
        order_by=input_assoc.c.input_data_id,
    )
    data = db.relationship("BaseData", foreign_keys=[data_id])

    __mapper_args__ = {"polymorphic_on": data_type}

    def __init__(self, name=None, project=None, description="", input_data=[]):
        self.project = project
        self.input_data = []
        for data in input_data:
            self.add_input(data)
        self.last_update = None
        self.name = name
        self.description = description

    def __deepcopy__(self, memo):
        other = type(self)()
        memo[id(self)] = other
        other.name = self.name
        other.description = self.description
        other.project = deepcopy(self.project, memo)
        other.input_data = [deepcopy(d, memo) for d in self.input_data]
        other.data = deepcopy(self.data, memo)
        return other

    def update_full(self):
        """Update project data.

        This makes sure that :attr:`ProjectData.last_update` field is equal
        to :attr:`ProjectData.modified_at` after update and database commit.
        So :meth:`ProjectData.is_outdated` method works as intended.

        :return:
        """
        super().update_modified()
        self.last_update = self.modified_at
        return super().save()

    def add_input(self, project_data):
        """Add input data.

        :param project_data:
        :raises ValueError: if `project_data` is from a different project
        """
        if project_data.project != self.project:
            raise ValueError("cannot add input data from other projects")
        self.input_data.append(project_data)

    def is_outdated(self):
        """Checks if input data are more recent than this object"""
        if self.last_update is None or self.last_update < self.modified_at:
            return True
        for data in self.input_data:
            if data.last_update is not None:
                if (
                    self.last_update is None
                    or data.last_update > self.last_update
                ):
                    return True
            if data.is_outdated():
                return True
        return False

    def get_attributes_list(self):
        """Returns the list of attributes of the features for this data.

        Returns:
            a dictionary with the following keys:
            "id", "name", "type", "statistics"
        """
        if self.data is not None:
            attr_list = []
            for attribute in self.data.attributes:
                attribute.load_statistics()
                attr_list.append(
                    {
                        "id": attribute.id,
                        "name": attribute.name,
                        "type": attribute.type.name,
                        "statistics": attribute.statistics,
                    }
                )
            return attr_list
        else:
            return []

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Return set of used input attributes of project data.

        :param data:
        :return: used input attributes
        """
        return set()

    def get_used_input_attributes(self) -> list[DataAttribute]:
        """Get used input attributes of project data if the user is authorized
        to access it.

        This list contains al inputl attributes used at least once in a
        processing model of the project data.

        Returns:
            Used input attributes.
        """
        return sorted(
            self._get_used_input_attributes(), key=lambda a: (a.data.id, a.id)
        )

    def _get_used_attributes(self) -> set[DataAttribute]:
        """Return set of used attributes of project data.

        This set contains all attributes used at least once in a processing
        model of the project data.

        :param data:
        :return: used input attributes
        """
        attributes = set()
        if self.data is None:
            return attributes
        for ref in self.output_data:
            attributes |= {
                attr
                for attr in ref._get_used_input_attributes()
                if attr.data_id == self.data.id
            }
        return attributes

    def get_used_attributes(self) -> list[DataAttribute]:
        """Get used attributes of project data if the user is authorized
        to access it.

        This list contains all attributes used at least once in a processing
        model of the downstream project data.

        Returns:
            Used attributes.
        """
        return sorted(self._get_used_attributes(), key=lambda a: a.id)

    def _explain_uptodate(self, feature: Feature) -> dict:
        """Explain uptodate feature values wrt the project data.

        :param feature: feature part of the project data
        :return: explanation as a dict
        """
        return {}

    def explain(self, feature: Feature) -> dict:
        """Explain feature values wrt the project data.

        :param feature:
        :raises ValueError: if `feature` does not belong to this project data
        :return: explanation as a dict
        """
        if self.is_outdated():
            return {}
        if feature.data_id != self.data_id:
            raise ValueError("feature does not belong to this project data")
        return self._explain_uptodate(feature)


class DataGeo(ProjectData):
    data = db.relationship("GeoData", foreign_keys=[ProjectData.data_id])

    __mapper_args__ = {"polymorphic_identity": DataType.geo_data}

    def __init__(self, data=None, data_id=None, **kwargs):
        if data is None and data_id is not None:
            data = GeoData.get_by_id(data_id)
        elif data is not None and data_id is not None:
            raise ValueError(
                "Only one of the parameters data or data_id must be passed"
            )
        if kwargs.get("name") is None and data is not None:
            kwargs["name"] = data.name
        super().__init__(**kwargs)
        self.data = data  # TODO: Verify the data is not None
        self.last_update = None

    def create(self):
        """Persist the object in the database, with a creation, modification
        and last updat edate"""
        self.created_at = datetime.now(timezone.utc)
        self.modified_at = self.created_at
        self.last_update = self.created_at
        return BaseModelMixin.create(self)

    def __repr__(self):
        return (
            "<DataGeo(name={s.name}, project_id={s.project_id}, "
            "data_id={s.data_id})>".format(s=self)
        )

    def add_input(self, project_data):
        """Add input (always fail).

        :param project_data:
        :raises TypeError: data doesn't accept inputs
        """
        raise TypeError(f"cannot add input in a {type(self).__name__} object")

    def is_outdated(self):
        """Checks if input data are more recent than this object"""
        return False


class ProjectGlobalData(ProjectData):
    data = db.relationship("GlobalData", foreign_keys=[ProjectData.data_id])

    __mapper_args__ = {"polymorphic_identity": DataType.global_data}

    def __init__(self, data=None, data_id=None, **kwargs):
        if data is None and data_id is not None:
            data = GlobalData.get_by_id(data_id)
        elif data is not None and data_id is not None:
            raise ValueError(
                "Only one of the parameters data or data_id must be passed"
            )
        if kwargs.get("name") is None and data is not None:
            kwargs["name"] = data.name
        super().__init__(**kwargs)
        self.data = data  # TODO: Verify the data is not None
        self.last_update = None

    def create(self):
        """Persist the object in the database, with a creation, modification
        and last updat edate"""
        self.created_at = datetime.now(timezone.utc)
        self.modified_at = self.created_at
        self.last_update = self.created_at
        return BaseModelMixin.create(self)

    def __repr__(self):
        return (
            "<ProjectGlobalData(name={s.name}, project_id={s.project_id}, "
            "data_id={s.data_id})>".format(s=self)
        )

    def add_input(self, project_data):
        """Add input (always fail).

        :param project_data:
        :raises TypeError: data doesn't accept inputs
        """
        raise TypeError(f"cannot add input in a {type(self).__name__} object")

    def is_outdated(self):
        """Checks if input data are more recent than this object"""
        return False


class DynamicData(ProjectData):
    @declared_attr
    def data(cls):
        return db.relationship(
            "GeneratedGeoData",
            single_parent=True,
            cascade="all,delete-orphan",
            foreign_keys=[ProjectData.data_id],
        )

    __mapper_args__ = {"polymorphic_abstract": True}


class DataStream(DynamicData):
    stream_id = db.Column(db.Integer, db.ForeignKey("base_data.id"))
    stream = db.relationship(BaseData, foreign_keys=[stream_id])
    classes = db.Column(ARRAY(db.Float, dimensions=2), nullable=True)
    start = db.Column(db.Float, nullable=True)
    step = db.Column(db.Float, nullable=True)
    stop = db.Column(db.Float, nullable=True)
    resolution = db.Column(db.Float, nullable=True)

    __mapper_args__ = {"polymorphic_identity": DataType.data_stream}

    def __init__(self, stream=None, stream_id=None, **kwargs):
        if stream is None and stream_id is not None:
            stream = BaseData.get_by_id(stream_id)
        elif stream is not None and stream_id is not None:
            raise ValueError(
                "Only one of the parameters stream or stream_id must be passed"
            )
        if not isinstance(stream, StreamGeoData):
            raise TypeError("'stream' field must reference a valid stream")
        if kwargs.get("name") is None and stream is not None:
            kwargs["name"] = stream.name
        if kwargs.get("description") is None and stream is not None:
            kwargs["description"] = stream.description
        if isinstance(stream, WMSGeoData):
            self.classes = kwargs.pop("classes", stream.classes)
            self.start = kwargs.pop("start", stream.start)
            self.step = kwargs.pop("step", stream.step)
            self.stop = kwargs.pop("stop", stream.stop)
            self.resolution = kwargs.pop("resolution", stream.resolution)
        super().__init__(**kwargs)
        self.stream = stream

    def __repr__(self):
        return (
            "<DataStream(name={s.name}, project_id={s.project_id}, "
            "stream_id={s.stream_id})>".format(s=self)
        )

    def __deepcopy__(self, memo):
        other = type(self)(self.stream)
        memo[id(self)] = other
        other.name = self.name
        other.description = self.description
        other.project = deepcopy(self.project, memo)
        other.input_data = [deepcopy(d, memo) for d in self.input_data]
        other.data = deepcopy(self.data, memo)
        if isinstance(self.stream, WMSGeoData):
            other.classes = self.classes
            other.start = self.start
            other.step = self.step
            other.stop = self.stop
            other.resolution = self.resolution
        return other

    def add_input(self, project_data):
        """Add input (always fail).

        :param project_data:
        :raises TypeError: data doesn't accept inputs
        """
        raise TypeError(f"cannot add input in a {type(self).__name__} object")


class DataGenerator(DynamicData):
    active_model_id = db.Column(
        db.Integer, db.ForeignKey("processing_model.id")
    )

    _processing_models = db.relationship(
        "ProcessingModel",
        cascade="all,delete-orphan",
        back_populates="data_generator",
        foreign_keys="[ProcessingModel.data_generator_id]",
        order_by="ProcessingModel.id",
    )
    _processing_model = db.relationship(
        "ProcessingModel",
        cascade="all",
        primaryjoin="DataGenerator.active_model_id == ProcessingModel.id",
        post_update=True,
    )

    __mapper_args__ = {"polymorphic_identity": DataType.generator}

    def __init__(self, processing_model=None, **kwargs):
        super().__init__(**kwargs)
        if self.processing_model is None:
            self.processing_model = processing_model
        self.data = None

    def __deepcopy__(self, memo):
        other = super().__deepcopy__(memo)
        other._processing_model = deepcopy(self._processing_model, memo)
        other._processing_models = [
            deepcopy(p, memo) for p in self._processing_models
        ]
        return other

    @property
    def processing_model(self):
        return self._processing_model

    @processing_model.setter
    def processing_model(self, model):
        self._processing_model = model
        if model not in self._processing_models:
            self._processing_models.append(model)

    def _get_used_input_attributes(self) -> set[DataAttribute]:
        """Return set of used input attributes of project data.

        This set contains all input attributes used at least once in a
        processing model of the project data.

        :return: used input attributes
        """
        attributes = set()
        for model in self._processing_models:
            attributes |= model._get_used_input_attributes()
        return attributes

    def _explain_uptodate(self, feature: Feature) -> dict:
        """Explain uptodate feature values wrt the project data.

        :param feature: feature part of the project data
        :return: explanation as a dict
        """
        return self.processing_model.explain(feature)


@db.event.listens_for(DataGenerator.name, "set", propagate=True)
def generator_name_set(target, value, oldvalue, initiator):
    """Event to modify the generated geo-data name when the project data name
    is modified"""
    if value != oldvalue and target.data is not None:
        target.data.name = value



from flask_restful import Resource

from ..schemas import UserListGetResponseSchema, UserSchema
from ..services.auth_service import token_required
from ..services.permission_service import get_all_users
from .utils import dump_data, with_response


class UserListAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=UserListGetResponseSchema,
        description="List of all users",
    )
    def get(self):
        """
        Return the list of all the users.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).

        .. :quickref: User; Get the list of users
        """
        users = get_all_users()
        schema = UserSchema(only=("id", "username"), many=True)
        return dump_data(schema, users=users)



"""Remove explainability field

Revision ID: 339b5d6198df
Revises: 141e705ab129
Create Date: 2025-04-08 11:38:17.264425

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "339b5d6198df"
down_revision = "141e705ab129"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "feature",
        sa.Column(
            "execution_artifact",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
    )

    conn = op.get_bind()
    nb_op = conn.execute(
        sa.text("SELECT count(id) FROM feature WHERE explainability != '[]'")
    ).fetchone()[0]
    res = conn.execute(
        sa.text(
            "SELECT id, explainability FROM feature "
            "WHERE explainability != '[]'"
        )
    )
    i = 0
    progress_bar(i, nb_op)
    for id, explainability in res:
        if not isinstance(explainability, dict):
            continue
        if explainability["model_type"] == "categories_rule":
            conn.execute(
                sa.text(
                    f"UPDATE feature SET execution_artifact = "
                    f"'{{\"rule\": \"{explainability['rule']}\"}}' "
                    f"WHERE id = {id}"
                )
            )
        i += 1
        progress_bar(i, nb_op)

    op.drop_column("feature", "explainability")


def downgrade():
    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "explainability",
                postgresql.JSONB(astext_type=sa.Text()),
                autoincrement=False,
                nullable=True,
            )
        )
        batch_op.drop_column("execution_artifact")



from celery.result import AsyncResult
from marshmallow import Schema, fields

from ..models import ProjectTaskModel, ProjectTaskType, User
from .base import BaseSchema, FieldMethod


class DefaultValueSchema(Schema):
    """Schema to use for creating or updating a default value"""

    name = fields.String(required=True)
    attribute = fields.String(required=True)
    default_value = fields.Field(required=True)


class DefaultValueDeletionSchema(Schema):
    """Schema to use when deleting a default value"""

    name = fields.String(required=True)
    attribute = fields.String(required=True)


class ProjectTaskSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = ProjectTaskModel
        # This schema is read-only so we do not need to change the write/read
        # fields
        # If this changes, we will need to define those and perhaps a creation
        # schema

    type = fields.Enum(ProjectTaskType, required=True)
    info = FieldMethod(fields.Dict(allow_none=True), "get_info")

    def get_info(self, obj):
        """Return the info if the task has the STARTED state."""
        result = AsyncResult(obj.task_id)
        if isinstance(result.info, dict):
            return result.info


class UserSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = User
        exclude = ("password_hash", "tokens", "permissions")
        # This schema is read-only so we do not need to change the write/read
        # fields
        # If this changes, we will need to define those and perhaps a creation
        # schema



import re

from marshmallow import ValidationError


def validate_name(name):
    """Validate a name against the python2 identifiers and keywords

    View https://docs.python.org/2/reference/lexical_analysis.html#identifiers

    Raises:
        ValidationError if the name is not valid
    """
    # python2 identifier definition, without allowing '_' as the first
    # character
    pattern = re.compile("[a-zA-Z][0-9a-zA-Z_]*")
    # python2 keywords and 'True', 'False' and 'None'
    reserved = [
        "and",
        "del",
        "from",
        "not",
        "while",
        "as",
        "elif",
        "global",
        "or",
        "with",
        "assert",
        "else",
        "if",
        "pass",
        "yield",
        "break",
        "except",
        "import",
        "print",
        "class",
        "exec",
        "in",
        "raise",
        "continue",
        "finally",
        "is",
        "return",
        "def",
        "for",
        "lambda",
        "try",
        "True",
        "False",
        "None",
    ]
    if name in reserved:
        raise ValidationError('"{}" is a reserved python name'.format(name))
    elif pattern.fullmatch(name) is None:
        raise ValidationError('"{}" is not a valid name'.format(name))



import shutil
import sys


def progress_bar(current, total):
    """A simple progress bar.

    :param current: progress
    :param total:
    """
    if total == 0:
        return
    # Get terminal width
    terminal_width = shutil.get_terminal_size((80, 20)).columns
    # Leave space for counters and brackets
    bar_length = terminal_width - len(str(total)) * 2 - 4

    progress = current / total
    bar = "=" * int(bar_length * progress) + "-" * (
        bar_length - int(bar_length * progress)
    )
    sys.stdout.write(f"\r[{bar}] {current}/{total}")
    if current == total:
        sys.stdout.write("\n")
    sys.stdout.flush()



import uuid
from datetime import datetime

from sqlalchemy.dialects.postgresql import UUID

from . import db
from .mixins import ModelMixin


class DataShare(ModelMixin, db.Model):
    __tablename__ = "data_share"

    data_id = db.Column(
        db.Integer, db.ForeignKey("base_data.id"), nullable=False
    )
    uid = db.Column(
        UUID, unique=True, nullable=False, default=lambda: str(uuid.uuid4())
    )
    expiration = db.Column(db.DateTime)
    expired = db.Column(db.Boolean, default=False)

    data = db.relationship("BaseData", back_populates="shares")

    @classmethod
    def get_by_uid(cls, uid):
        return cls.query.filter_by(uid=uid).first()

    def is_expired(self):
        if self.expiration is None:
            date_expired = False
        else:
            date_expired = self.expiration < datetime.now()
        return date_expired or self.expired



import time
from collections import defaultdict

import numpy as np
from flask import current_app as app
from geoalchemy2 import Geography, Geometry
from geoalchemy2.shape import from_shape, to_shape
from pandas import DataFrame
from rtree import index as rindex
from shapely.errors import TopologicalError
from shapely.geometry import MultiPolygon, Polygon
from shapely.geometry.base import BaseMultipartGeometry
from shapely.ops import unary_union
from shapely.prepared import prep

from ..exceptions import ProcessingError
from ..services import genetic_algorithm_service as ga_service
from ..services.geo_data_loading_service import post_load_normalization
from . import db
from .genetic_algorithm import GeneticAlgorithm
from .geo_data import (
    DataAttributeNominal,
    DataAttributeOrdinal,
    DataAttributeQuantitative,
    GeneratedGeoData,
    GeoFeature,
    GlobalData,
)
from .processing_model_utils import evaluate_continuous_rule
from .processing_models import (
    ContinuousRule,
    DiscreteRules,
    DissolveAdjacentModel,
    GeoBuffer,
    MergeOverlapModel,
    MRSort,
    PrefDefaultValues,
    WeightedSum,
    ZonePropositionGenerator,
)


def check_geom_type(geom):
    result = None
    if geom.type == "Polygon":
        result = geom
    elif (
        geom.type in ("MultiPolygon", "GeometryCollection")
        and not geom.is_empty
    ):
        geoms = []
        for g in geom.geoms:
            new_geom = check_geom_type(g)
            if new_geom is not None and not new_geom.is_empty:
                geoms.append(new_geom)
        if len(geoms) == 1:
            result = geoms[0]
        elif len(geoms) > 1:
            result = MultiPolygon(geoms)
    if result is not None and result.area <= 1e-08:
        result = None
    if result is not None:
        return result.buffer(0)
    else:
        return Polygon()


class GeometryFeature:
    def __init__(self, ids, geometry):
        self.ids = frozenset(ids)
        self.geometry = geometry

    @classmethod
    def from_dict(cls, value):
        from shapely.wkt import loads

        return cls(ids=value["ids"], geometry=loads(value["geometry"]))

    def to_dict(self):
        from shapely.wkt import dumps

        return {"ids": list(self.ids), "geometry": dumps(self.geometry)}

    def intersection(self, other):
        ids = self.ids | other.ids
        try:
            geometry = self.geometry.intersection(other.geometry)
            return GeometryFeature(ids=ids, geometry=geometry)
        except TopologicalError:
            app.logger.warning(
                "Topological error in the intersection of {} and {}".format(
                    self.ids, other.ids
                )
            )
            return GeometryFeature(ids=ids, geometry=Polygon())

    def difference(self, other):
        ids = self.ids
        try:
            geometry = self.geometry.difference(other.geometry)
            return GeometryFeature(ids=ids, geometry=geometry)
        except TopologicalError:
            app.logger.warning(
                "Topological error in the intersection of {} and {}".format(
                    self.ids, other.ids
                )
            )
            return GeometryFeature(ids=ids, geometry=Polygon())

    def difference_union(self, geoms):
        ids = self.ids
        # FIXME: Fix the errors instead of simply ignoring them
        try:
            geom_union = unary_union([g.geometry for g in geoms])
        except ValueError:
            app.logger.warning(
                "Error in the union of geometries of {} and {}".format(
                    ids, [g.ids for g in geoms]
                )
            )
            return GeometryFeature(ids=ids, geometry=Polygon())
        if geom_union.is_empty:
            return GeometryFeature(ids=ids, geometry=self.geometry)
        try:
            geometry = self.geometry.difference(geom_union)
        except TopologicalError:
            app.logger.warning(
                "Topological error in the difference of {} and {}".format(
                    ids, [g.ids for g in geoms]
                )
            )
            return GeometryFeature(ids=ids, geometry=Polygon())
        return GeometryFeature(ids=ids, geometry=geometry)

    def split_multi_geometries(self) -> list["GeometryFeature"]:
        match self.geometry:
            case BaseMultipartGeometry():
                res = []
                for g in self.geometry.geoms:
                    res.append(GeometryFeature(self.ids, g))
                return res
            case _:
                return [self]

    def is_valid(self):
        return self.geometry.is_valid

    def is_empty(self):
        return self.geometry.is_empty

    def intersects(self, other):
        return self.geometry.intersects(other.geometry)

    def touches(self, other):
        return self.geometry.touches(other.geometry)


class ModelProcessingService:
    def decompose(self, geometries_1, geometries_2, r_tree1, feedback):
        geoms = []
        rtree_final = rindex.Index()

        def add_geometry(geom):
            g = check_geom_type(geom.geometry)
            if g is None or g.is_empty:
                return
            geom.geometry = g
            for _geom in geom.split_multi_geometries():
                rtree_final.insert(len(geoms), _geom.geometry.bounds)
                geoms.append(_geom)

        count = 0
        total = len(geometries_1) + len(geometries_2)

        r_tree2 = rindex.Index()
        for id2, geom2 in enumerate(geometries_2):
            r_tree2.insert(id2, geom2.geometry.bounds)
            prepared_geom2 = prep(geom2.geometry)
            geoms1 = []
            for id1 in r_tree1.intersection(geom2.geometry.bounds):
                if prepared_geom2.intersects(geometries_1[id1].geometry):
                    g = geom2.intersection(geometries_1[id1])
                    add_geometry(g)
                    geoms1.append(geometries_1[id1])
            g = geom2.difference_union(geoms1)
            add_geometry(g)
            count += 1
            feedback.set_progress(count, total)

        for geom1 in geometries_1:
            prepared_geom1 = prep(geom1.geometry)
            geoms2 = [
                geometries_2[id2]
                for id2 in r_tree2.intersection(geom1.geometry.bounds)
                if prepared_geom1.intersects(geometries_2[id2].geometry)
            ]
            g = geom1.difference_union(geoms2)
            add_geometry(g)
            count += 1
            feedback.set_progress(count, total)

        return geoms, rtree_final

    def buffer_geometries(self, distance, geometries):
        """Create a buffer around the geometries.

        The geometries are cast to a geography type in PostGIS, allowing to use
        a distance in metres for the buffer.

        Args:
            distance: The distance of the buffer.
            geometries (iterable): Iterable of GeometryFeature objects

        Returns:
            A new list of GeometryFeature objects with a buffer around the
            geometries.
        """
        q = db.select(
            db.cast(
                db.func.st_buffer(
                    db.cast(db.column("geom"), Geography), distance
                ),
                Geometry,
            )
        )
        q = q.select_from(
            db.func.unnest([g.geometry.wkt for g in geometries]).alias("geom")
        )
        result = db.session.execute(q).fetchall()
        buffered = [
            GeometryFeature(g.ids, to_shape(res[0]))
            for g, res in zip(geometries, result)
        ]
        return buffered

    def _get_default_values(self, model):
        values = {}
        for d in model.input_data:
            if d.data is not None:
                attrs = d.data.attributes
                attributes = {}
                for attr in attrs:
                    value = None
                    if isinstance(model, PrefDefaultValues):
                        # replace by default values
                        value = model.get_default_value(attr)
                    attributes[attr.name] = value
                values[d.name] = {
                    "properties": {"area": None},
                    "attributes": attributes,
                }
        return values

    def _get_sandbox_data(self, model, geometries, input_features):
        features_ids = frozenset.union(*(g.ids for g in geometries))
        q = db.session.query(
            GeoFeature, db.func.st_area(db.cast(GeoFeature.geom, Geography))
        )
        q = q.filter(GeoFeature.id.in_(list(features_ids)))
        q_result = q.all()

        features_data = {}
        project_data_names = {
            d.data.id: d.name for d in model.input_data if d.data is not None
        }
        default_values = self._get_default_values(model)

        for f, area in q_result:
            properties = {"area": area}
            attributes = {p.attribute.name: p.value for p in f.properties}
            if isinstance(model, ContinuousRule):
                # replace None by default values
                for p in f.properties:
                    if p.value is not None and not isinstance(p.value, float):
                        continue
                    if (
                        isinstance(p.value, float)
                        and p.value == p.value
                        and p.value != float("inf")
                        and p.value != -float("inf")
                    ):
                        continue
                    attr = p.attribute
                    attributes[attr.name] = model.get_default_value(attr)
            features_data[f.id] = {
                "name": project_data_names[f.data.id],
                "data": {"properties": properties, "attributes": attributes},
            }

        for fId in features_ids - set(features_data.keys()):
            f = input_features[fId]
            properties = {"area": None}
            attributes = {p.attribute.name: p.value for p in f.properties}
            if isinstance(model, ContinuousRule):
                # replace None by default values
                for p in f.properties:
                    if p.value is not None and not isinstance(p.value, float):
                        continue
                    if (
                        isinstance(p.value, float)
                        and p.value == p.value
                        and p.value != float("inf")
                        and p.value != -float("inf")
                    ):
                        continue
                    attr = p.attribute
                    attributes[attr.name] = model.get_default_value(attr)
            features_data[f.id] = {
                "name": project_data_names[f.data.id],
                "data": {"properties": properties, "attributes": attributes},
            }

        process_list = []
        areas = self.get_geometries_area(geometries)
        for g, area in zip(geometries, areas):
            data = {}
            data[model.data_generator.name] = {
                "properties": {"area": area},
                "attributes": {},
            }
            for f_id in g.ids:
                feature = features_data[f_id]
                data[feature["name"]] = feature["data"]
            for k, v in default_values.items():
                if k not in data:
                    data[k] = v
            process_list.append(data)

        return process_list

    def _get_attributes_stats(self, model):
        """
        Return the statistics of the input data attributes.

        The statistics are returned as a dictionary containing the stats in the
        form {'data name': {'attribute name': {'stat1': value, 'stat2': value,
        ...}}}

        The returned statistics are 'min', 'max', 'mean', 'std' and 'count' if
        they already exist in the attribute.
        Percentiles are also returned from 'percentile_0' to 'percentile_100'.
        """
        input_attributes = []
        for data in model.input_data:
            if data.data is not None:
                data.data.load_properties()
                input_attributes += data.data.attributes
        stats = defaultdict(dict)
        for attr in input_attributes:
            keys = ["min", "max", "mean", "std", "count"]
            if attr.statistics is not None:
                attr_stat = {}
                for stat_key, stat_val in attr.statistics.items():
                    if stat_key in keys:
                        keys.remove(stat_key)
                        attr_stat[stat_key] = stat_val
                try:
                    df = DataFrame([v.value for v in attr.values])
                    series = df.quantile(np.arange(0, 1.01, 0.01))[0]
                    for k, v in series.items():
                        attr_stat["percentile_{}".format(int(k * 100))] = v
                except (TypeError, ValueError):
                    app.logger.info(
                        "The attribute is not a number, percentiles ignored"
                    )
                stats[attr.data.name][attr.name] = attr_stat
        return dict(stats) if len(stats) > 0 else None

    def _get_inference_data(self, model, geometries, features):
        attributes = []
        for d in model.input_data:
            if d.data is not None:
                attributes += d.data.attributes
        data_list = []
        for geometry in geometries:
            data = {a: None for a in attributes}
            for f_id in geometry.ids:
                feature = features.get(f_id)
                for value in feature.properties:
                    data[value.attribute] = value.value
            data_list.append(data)
        return data_list

    def _get_result_attributes(self, result_list, model):
        old_attrs = {}
        if model.data_generator.data is not None:
            old_attrs = {
                a.name: a for a in model.data_generator.data.attributes
            }
        is_quantitative = (
            lambda v: isinstance(v, int) or isinstance(v, float) or v is None
        )
        attrs = {}
        for properties in result_list:
            for name, value in properties.items():
                if name not in attrs:
                    if is_quantitative(value):
                        new = DataAttributeQuantitative(name=name)
                    else:
                        new = DataAttributeNominal(name=name)
                    if name in old_attrs:
                        old = old_attrs[name]
                        if new.type == old.type:
                            new = old
                    attrs[name] = new
                if isinstance(
                    attrs[name], DataAttributeQuantitative
                ) and not is_quantitative(value):
                    attrs[name] = DataAttributeNominal(name=name)
        return attrs

    def get_decomposed_data(self, model):
        # TODO: Find a cleaner way to get the decomposed data values
        geometries, features = self._process_regular_geometries(model)
        return self._get_inference_data(model, geometries, features)

    def _process_rule(self, model, feedback):
        """Process a rule model."""
        geometries, features = self._process_regular_geometries(
            model, feedback
        )
        data_json = self._get_sandbox_data(model, geometries, features)
        data_stats = self._get_attributes_stats(model)
        result_list = evaluate_continuous_rule(
            data=data_json, rule=model.rule, stats=data_stats
        )
        attrs = self._get_result_attributes(result_list, model)
        new_features = []
        for result, g in zip(result_list, geometries):
            values = []
            input_features = [features[fid] for fid in g.ids]
            new_feature = GeoFeature(
                geom=from_shape(g.geometry, srid=4326),
                input_features=input_features,
            )
            for name, value in result.items():
                # Ignore null values
                if value is None:
                    continue
                attr = attrs[name]
                cls = attr.get_value_class()
                values.append(
                    cls(value=value, attribute=attr, feature=new_feature)
                )
            # If the feature has values
            if len(values) > 0:
                new_features.append(new_feature)
        return new_features

    def _process_model_buffer(self, model, feedback):
        """Copy the attributes from the geometries."""
        geometries, features_dict = self._process_buffer_geometry(
            model, feedback
        )
        geo_data = None
        for f in features_dict.values():
            if geo_data is None:
                geo_data = f.data
            else:
                assert geo_data is f.data
        old_attrs = {}
        if model.data_generator.data is not None:
            old_attrs = {
                a.name: a for a in model.data_generator.data.attributes
            }
        attrs = {}
        for a in geo_data.attributes:
            old_attr = old_attrs.get(a.name)
            attrs[a.id] = old_attr if a.same_as(old_attr) else a.new_copy()

        new_features = []
        for g in geometries:
            f_id = list(g.ids)[0]  # Copy attributes from the first geometry
            old_feature = features_dict[f_id]
            new_feature = GeoFeature(
                geom=from_shape(g.geometry, srid=4326),
                input_features=[old_feature],
            )
            new_props = []
            for v in old_feature.properties:
                cls = attrs[v.attribute_id].get_value_class()
                val = cls(
                    value=v.value,
                    attribute=attrs[v.attribute_id],
                    feature=new_feature,
                )
                new_props.append(val)
            new_features.append(new_feature)
        return new_features

    def _process_criteria_rules(self, model, feedback):
        """Process criteria rules.

        A category will be assigned for each of the resulting features, in a
        'category' attribute.

        Returns:
            A list of Feature objects.
        """
        geometries, features = self._process_regular_geometries(
            model, feedback
        )
        data_json = self._get_sandbox_data(model, geometries, features)
        data_stats = self._get_attributes_stats(model)
        results = []
        for category in model.categories:
            evaluation = category.evaluate_data(data_json, data_stats)
            results.append((category.name, category.rules, evaluation))
        final = np.ma.masked_all_like(data_json)
        final_rule = np.ma.masked_all_like(data_json)
        for categ, rules, evaluation in results:
            for rule, row in zip(rules, evaluation):
                mask = np.where(row & final.mask)
                final[mask] = categ
                final_rule[mask] = rule
        if np.ma.is_masked(final):
            raise ProcessingError("No rule validated")
        # TODO: Send a warning if several categories are valid
        # (r.count(True) > 1)
        order = [c.name for c in model.categories]
        attribute = DataAttributeOrdinal(name="category", order=order)
        if model.data_generator.data is not None:
            for attr in model.data_generator.data.attributes:
                if attribute.same_as(attr):
                    attribute = attr
                    break
        new_features = []
        for categ, rule, g, data in zip(
            final, final_rule, geometries, data_json
        ):
            input_features = [features[fid] for fid in g.ids]
            new_feature = GeoFeature(
                geom=from_shape(g.geometry, srid=4326),
                execution_artifact={"rule": rule},
                input_features=input_features,
            )
            cls = attribute.get_value_class()
            cls(value=categ, attribute=attribute, feature=new_feature)
            new_features.append(new_feature)
        return new_features

    def _process_mr_sort(self, model, feedback):
        """Process using the MR-Sort algorithm.

        Each resulting feature will be assigned a category by the MR-Sort
        algorithm, in a 'category' attribute.

        Returns:
            A list of Feature objects.
        """
        geometries, features = self._process_regular_geometries(
            model, feedback
        )
        data_json = self._get_sandbox_data(model, geometries, features)
        new_features = []
        order = [c.name for c in model.categories]
        attribute = DataAttributeOrdinal(name="category", order=order)
        if model.data_generator.data is not None:
            for attr in model.data_generator.data.attributes:
                if attribute.same_as(attr):
                    attribute = attr
                    break
        profiles = {}
        for criterion in model.criteria:
            for index, value in enumerate(criterion.profiles):
                if index not in profiles:
                    profiles[index] = []
                profiles[index].append(
                    (
                        criterion.attribute.data.name,
                        criterion.attribute.name,
                        value,
                    )
                )
        for data, g in zip(data_json, geometries):
            attributes = []
            input_features = [features[fid] for fid in g.ids]
            for d_name, props in data.items():
                if d_name == model.data_generator.name:
                    continue
                for prop, val in props["attributes"].items():
                    attributes.append((d_name, prop, val))
            category = model.check_category(attributes)
            new_feature = GeoFeature(
                geom=from_shape(g.geometry, srid=4326),
                input_features=input_features,
            )
            cls = attribute.get_value_class()
            props = [
                cls(
                    value=category.name,
                    attribute=attribute,
                    feature=new_feature,
                )
            ]
            new_features.append(new_feature)
        return new_features

    def _process_weighted_sum(self, model, feedback):
        geometries, features = self._process_regular_geometries(
            model, feedback
        )
        new_features = []
        attribute = DataAttributeQuantitative(name="result")
        if model.data_generator.data is not None:
            for attr in model.data_generator.data.attributes:
                if attribute.same_as(attr):
                    attribute = attr
                    break
        default_values = {
            d.attribute_id: d.value for d in model.default_values
        }
        for geom in geometries:
            value = None
            input_features = [features[fid] for fid in geom.ids]
            vals = {
                p.attribute_id: p.value
                for f in input_features
                for p in f.properties
            }
            for operand in model.operands:
                new_val = vals.get(operand.attribute_id)
                if new_val is None:
                    new_val = default_values.get(operand.attribute_id)
                if new_val is not None:
                    new_val = operand.weight * new_val
                    value = new_val if value is None else value + new_val
                else:
                    # Ignore the feature if there is no default value
                    value = None
                    break
            if value is not None:
                new_feature = GeoFeature(
                    geom=from_shape(geom.geometry, srid=4326),
                    input_features=input_features,
                )
                cls = attribute.get_value_class()
                props = [  # noqa: F841 (those props are persisted by ORM)
                    cls(value=value, attribute=attribute, feature=new_feature)
                ]
                new_features.append(new_feature)
        return new_features

    def _decompose_features(self, geometries: list[GeometryFeature], feedback):
        rtree = rindex.Index()
        geoms: dict[int, GeometryFeature] = {}

        def add_geometry(geom, id_=None):
            if id_ is not None and id_ in geoms:
                rtree.delete(id_, geoms[id_].geometry.bounds)
            ids_ = [] if id_ is None else [id_]
            # Split multi-part geometries on insertion/edition
            for _geom in geom.split_multi_geometries():
                id_ = ids_.pop() if len(ids_) > 0 else len(geoms)
                rtree.insert(id_, _geom.geometry.bounds)
                geoms[id_] = _geom

        for g in geometries:
            add_geometry(g)

        res = []
        id1 = -1
        feedback.set_message("Decomposing overlaps")
        while id1 < len(geoms) - 1:
            id1 += 1
            geom1 = geoms[id1]
            if geom1 is None:
                feedback.set_progress(id1, len(geoms))
                continue
            rtree.delete(id1, geom1.geometry.bounds)
            for id2 in rtree.intersection(geom1.geometry.bounds):
                geom2 = geoms[id2]
                try:
                    if not geom1.intersects(geom2):
                        continue

                    intersection = geom1.intersection(geom2)
                    intersection.geometry = check_geom_type(
                        intersection.geometry
                    )
                except TopologicalError:
                    intersection = None
                    app.logger.warning("Intersection error in the overlap")
                if intersection is not None and not intersection.is_empty():
                    add_geometry(intersection)
                else:
                    continue

                try:
                    g2diff = geom2.difference(geom1)
                    g2diff.geometry = check_geom_type(g2diff.geometry)
                except TopologicalError:
                    g2diff = None
                    app.logger.warning("Difference error in the overlap")
                if g2diff is not None and not g2diff.is_empty():
                    add_geometry(g2diff, id2)
                else:
                    rtree.delete(id2, geom2.geometry.bounds)
                    geoms[id2] = None

                try:
                    g1diff = geom1.difference(geom2)
                    g1diff.geometry = check_geom_type(g1diff.geometry)
                except TopologicalError:
                    g1diff = None
                    app.logger.warning("Difference error in the overlap")
                if g1diff is not None and not g1diff.is_empty():
                    geom1 = g1diff
                else:
                    geom1 = None
                    break
            if geom1 is not None:
                for geom in geom1.split_multi_geometries():
                    res.append(geom)
            feedback.set_progress(id1, len(geoms))
        feedback.forget_message()
        return res

    def _process_buffer_geometry(self, model, feedback):
        """Create a buffer around the geometries.

        Returns:
            A list of GeometryFeature and the dict of input features.
        """
        # Prepare the extent
        extent = to_shape(model.data_generator.project.extent)
        extent_prep = prep(extent)
        # Check the input data is correct
        if len(model.input_data) != 1:
            raise ProcessingError("Incorrect number of input data")
        input_data = model.input_data[0]
        if input_data.data is None:
            raise ProcessingError("No geometries for the input data")
        # Convert the features to Shapely geometries
        geoms = []
        features = {}
        for feature in input_data.data.features:
            geom = to_shape(feature.geom)
            geoms.append(GeometryFeature(ids=[feature.id], geometry=geom))
            features[feature.id] = feature
        # Buffer
        geometries = self.buffer_geometries(model.radius, geoms)
        # Filter the geometries that intersect with the extent
        filtered = []
        for geom in geometries:
            if extent_prep.intersects(geom.geometry):
                if model.cut_to_extent:
                    geom.geometry = geom.geometry.intersection(extent)
                filtered.append(geom)
        return filtered, features

    def _process_regular_geometries(self, model, feedback=None):
        """Decompose the geometries.

        .. todo:: rename to something like `flatten_geometry_sets`

        Returns:
            A list of GeometryFeature, and the dict of features
        """
        send_feedback = (
            (lambda *args, **kwargs: None)
            if feedback is None
            else feedback.set_message
        )
        # Prepare the extent
        extent = to_shape(model.data_generator.project.extent)
        extent_prep = prep(extent)
        # Convert the features to Shapely geometries and cut them to the extent
        # if needed
        input_geometries = []
        input_features = {}
        data_list = (i.data for i in model.input_data if i.data is not None)
        global_features_ids = []
        for data in data_list:
            if isinstance(data, GlobalData) and data.feature:
                input_features[data.feature.id] = data.feature
                global_features_ids.append(data.feature.id)
                continue
            geoms = []
            for feature in data.features:
                geom = to_shape(feature.geom)
                if extent_prep.intersects(geom) and not geom.is_empty:
                    if model.cut_to_extent:
                        geom = geom.intersection(extent)
                    geoms.append(
                        GeometryFeature(ids=[feature.id], geometry=geom)
                    )
                    input_features[feature.id] = feature
            input_geometries.append(geoms)
        # Check the input data exists
        if len(input_geometries) == 0:
            raise ProcessingError("No input data exist")
        # Decompose the geometries
        geometries = input_geometries[0]
        rtree = rindex.Index()
        for i, g in enumerate(geometries):
            rtree.insert(i, g.geometry.bounds)
        for i, geoms in enumerate(input_geometries[1:]):
            send_feedback(
                "Decomposing geometries ({}/{})".format(
                    i + 1, len(input_geometries) - 1
                )
            )
            geometries, rtree = self.decompose(
                geometries, geoms, rtree, feedback=feedback
            )
        # Add global input features
        geometries = [
            GeometryFeature(
                list(geom.ids) + global_features_ids, geom.geometry
            )
            for geom in geometries
        ]

        return geometries, input_features

    def _process_overlap(self, model, feedback):
        # Prepare the extent
        extent = to_shape(model.data_generator.project.extent)
        extent_prep = prep(extent)
        # Check the input data is correct
        if len(model.input_data) != 1:
            raise ProcessingError("Incorrect number of input data")
        input_data = model.input_data[0]
        if input_data.data is None:
            raise ProcessingError("No geometries for the input data")
        # Convert the features to Shapely geometries
        # Filter geometries that intersects with the extend
        geoms = []
        features = {}
        for feature in input_data.data.features:
            geom = to_shape(feature.geom)
            if extent_prep.intersects(geom):
                if model.cut_to_extent:
                    geom = geom.intersection(extent)
                geoms.append(GeometryFeature(ids=[feature.id], geometry=geom))
                features[feature.id] = (geom, feature)

        # Duplicate all input attributes
        old_attrs = {}
        if model.data_generator.data is not None:
            old_attrs = {
                a.name: a for a in model.data_generator.data.attributes
            }
        attrs = {}
        for a in input_data.data.attributes:
            old_attr = old_attrs.get(a.name)
            attrs[a.id] = old_attr if a.same_as(old_attr) else a.new_copy()

        new_features = []
        geoms = self._decompose_features(geoms, feedback=feedback)
        attr_values = defaultdict(list)
        for geom in geoms:
            input_features = []
            new_feature = GeoFeature(geom=from_shape(geom.geometry, srid=4326))
            g1 = prep(geom.geometry)
            tmp_values = defaultdict(list)
            for id2 in geom.ids:
                # Each geometry was split multiple times
                # We cannot be sure that all input feature ids intersect with
                # resulting geometry
                geom2, feature2 = features[id2]
                if g1.intersects(geom2):
                    input_features.append(feature2)
                    for value in feature2.properties:
                        if value is not None:
                            attr = attrs[value.attribute_id]
                            tmp_values[attr].append(value)
            new_feature.input_features = input_features
            values = []
            for attr, vals in tmp_values.items():
                if len(vals) >= 1:
                    try:
                        val = attr.get_value_class()(
                            value=model.get_overlap_value(vals),
                            attribute=attr,
                            feature=new_feature,
                        )
                        attr_values[attr].append(val)
                    except TypeError:
                        raise ProcessingError(
                            "Error overlapping the values types."
                        )
            new_features.append(new_feature)
        for attr, values in attr_values.items():
            attr.values = values
        return new_features

    def process_model(self, model, feedback):
        """Process the model on the geometries.

        Args:
            model: A processing model.
            geometries (list): A list of GeometryFeature objects.

        Returns:
            A list of Feature objects.
        """
        features = []
        feedback.set_message("Processing model")
        if isinstance(model, GeoBuffer):
            features = self._process_model_buffer(model, feedback)
        elif isinstance(model, ContinuousRule):
            features = self._process_rule(model, feedback)
        elif isinstance(model, DiscreteRules):
            features = self._process_criteria_rules(model, feedback)
        elif isinstance(model, MRSort):
            features = self._process_mr_sort(model, feedback)
        elif isinstance(model, WeightedSum):
            features = self._process_weighted_sum(model, feedback)
        elif isinstance(model, MergeOverlapModel):
            features = self._process_overlap(model, feedback)
        elif isinstance(model, DissolveAdjacentModel):
            features = self._process_dissolve(model, feedback)
        elif isinstance(model, ZonePropositionGenerator):
            features = self._process_zone_proposition(model, feedback)
        else:
            raise ValueError
        return features

    def set_project_data_features(
        self, project_data, features, filter_none=True
    ):
        """Save the features.

        If a geo data already exists, the old features are replaced by the
        features given in arguements.
        If no geo data exists, a new one is created using the features.

        Args:
            project_data (ProjectData)
            features (list of GeoFeature)
            filter_none (bool): Remove the features with only None values
                (default: True)
        """
        if filter_none:
            # Remove features without values
            features_original = list(features)
            features = []
            for feat in features_original:
                if any((v.value is not None for v in feat.properties)):
                    features.append(feat)

        # Remove features with empty geometries
        features = list(
            filter(lambda f: to_shape(f.geom).is_empty is False, features)
        )

        # List the data attributes
        attrs = []
        for f in features:
            for p in f.properties:
                if p.attribute not in attrs:
                    attrs.append(p.attribute)

        if project_data.data is None:
            # Create a new data...
            project_data.data = GeneratedGeoData(
                name=project_data.name,
                features=features,
                attributes=attrs,
            )
        else:
            # ...or modify the old one
            data = project_data.data
            data.name = project_data.name
            data.features = features
            data.extent = None
            data.attributes = attrs
        project_data.data.update()
        post_load_normalization(project_data.data)
        project_data.update_full()

    def get_geometries_area(self, geometries):
        """Return the areas of the geometries.

        Calculates the area in PostGIS by casting the geometries to Geography
        data, allowing to have the area in square metres.

        Args:
            geometries (list): List of GeometryFeature objects.
        """
        q = db.select(db.func.st_area(db.cast(db.column("geom"), Geography)))
        q = q.select_from(
            db.func.unnest([g.geometry.wkt for g in geometries]).alias("geom")
        )
        q_result = db.session.execute(q).fetchall()
        areas = [r[0] for r in q_result]
        return areas

    def _process_dissolve(self, model, feedback):
        """Dissolve the adjacent geometries with the same attributes.

        Does not do anything if the dissolve_adjacent attribute of the model is
        not set. Otherwise, if two geometries touch each other and have the
        same values for their attributes, they will be merged in a single
        geometry.
        """
        # Prepare the extent
        extent = to_shape(model.data_generator.project.extent)
        extent_prep = prep(extent)
        # Check the input data is correct
        if len(model.input_data) != 1:
            raise ProcessingError("Incorrect number of input data")
        input_data = model.input_data[0]
        if input_data.data is None:
            raise ProcessingError("No geometries for the input data")

        # Duplicate all input attributes
        old_attrs = {}
        if model.data_generator.data is not None:
            old_attrs = {
                a.name: a for a in model.data_generator.data.attributes
            }
        attrs = {}
        for a in input_data.data.attributes:
            old_attr = old_attrs.get(a.name)
            attrs[a.id] = old_attr if a.same_as(old_attr) else a.new_copy()

        # Convert the features to Shapely geometries
        # Filter geometries that intersects with the extend
        # Group the geometries when the features have identical attributes
        # values
        # keys are a tuple of (attribute, value) and values are lists of
        # shapely geometries
        grouped = defaultdict(list)
        grouped_features = defaultdict(list)
        for feature in input_data.data.features:
            geom = to_shape(feature.geom)
            if extent_prep.intersects(geom):
                if model.cut_to_extent:
                    geom = geom.intersection(extent)
                f_key = tuple(
                    (prop.attribute_id, prop.value)
                    for prop in feature.properties
                )
                grouped[f_key].append(geom)
                grouped_features[f_key].append(feature)

        # Create the new features
        new_features = []
        attr_values = defaultdict(list)
        for key, geoms in grouped.items():
            # Union of each group of geometries
            union = unary_union(geoms)
            new_geoms = (
                list(union.geoms)
                if isinstance(union, BaseMultipartGeometry)
                else [union]
            )
            for new_geom in new_geoms:
                new_prep_geom = prep(new_geom)
                input_features = []
                for feature, geom in zip(grouped_features[key], geoms):
                    if new_prep_geom.intersects(geom):
                        input_features.append(feature)
                new_feature = GeoFeature(
                    geom=from_shape(new_geom, srid=4326),
                    input_features=input_features,
                )
                for attr_id, value in key:
                    attr = attrs[attr_id]
                    cls = attr.get_value_class()
                    val = cls(
                        value=value,
                        attribute=attr,
                        feature=new_feature,
                    )
                    attr_values[attr].append(val)
                new_features.append(new_feature)
        for attr, values in attr_values.items():
            attr.values = values
        return new_features

    def _process_zone_proposition(self, model, feedback):
        """Make recommendations for zones.

        :param model:
        :param feedback:
        :raises ProcessingError: if there is not exactly 1 input_data
        :raises ProcessingError: if input_data has no geometry
        :raises ProcessingError: if no end-condition is set for model
        :return: generated zones as features
        """
        # Check the input data is correct
        if len(model.input_data) != 1:
            raise ProcessingError("Incorrect number of input data")
        input_data = model.input_data[0]
        if input_data.data is None:
            raise ProcessingError("No geometries for the input data")
        iterations, duration = model.iterations, model.duration
        if iterations is None and duration is None:
            raise ProcessingError("No end condition for zone proposition run")
        ga_ctx = ga_service.ga_context_from_decision_map(
            input_data,
            model.geo_size,
            proj=(
                "+proj=cea +lon_0=0 +lat_ts=45 +x_0=0 +y_0=0 +ellps=WGS84 "
                "+datum=WGS84 +units=m +no_defs"
            ),
        )
        ga = GeneticAlgorithm(ga_ctx, **model.ga_params())
        time_start = time.time()
        iteration_continue, duration_continue = True, True
        while iteration_continue and duration_continue:
            ga.iterate()
            current_duration = time.time() - time_start
            iteration_continue = (
                ga.iteration < iterations if iterations is not None else True
            )
            duration_continue = (
                current_duration < duration if duration is not None else True
            )
            if iterations is not None and duration is None:
                progress_count = ga.iteration
                progress_total = iterations
            elif iterations is None and duration is not None:
                progress_count = current_duration
                progress_total = duration
            elif ga.iteration / iterations >= current_duration / duration:
                progress_count = ga.iteration
                progress_total = iterations
            else:
                progress_count = current_duration
                progress_total = duration
            feedback.set_progress(progress_count, progress_total)
        features = []
        population = ga.filter_best()
        attribute = DataAttributeQuantitative(name="fitness")
        value_cls = attribute.get_value_class()
        for individual in population.individuals:
            value = value_cls(value=individual.fitness, attribute=attribute)
            feature = GeoFeature(
                from_shape(
                    individual.as_geometry(proj="epsg:4326"), srid=4326
                ),
                properties=[value],
                input_features=individual.inputs,
                execution_artifact=individual.execution_artifact,
            )
            features.append(feature)
        return features

    def dissolve_adjacent_features(self, data_id):
        """Dissolve the adjacent features with the same attributes values."""
        q = "SELECT dissolve_adjacent(:data_id)"
        keys = {"data_id": data_id}
        db.session.execute(q, keys)



from .geo_data import (
    DataService,
    GeoDataService,
    GlobalDataService,
    StreamGeoDataService,
)
from .processing_model import ProcessingModelService
from .project import ProjectService
from .project_data import ProjectDataService

project_service = ProjectService()
project_data_service = ProjectDataService()
data_service = DataService()
geo_data_service = GeoDataService()
geo_data_stream_service = StreamGeoDataService()
processing_model_service = ProcessingModelService()
get_global_data_service = GlobalDataService()



from marshmallow import fields
from marshmallow_sqlalchemy import field_for

from ..models.mrsort_inference import MRSortInferenceAlternative
from ..models.processing_models import MRSort, MRSortCriterion
from .base import BaseSchema, FieldMethod
from .geo_data import BaseDataAttributeSchema
from .processing_models import DiscreteCategorySchema


class MRSortInferenceAlternativeSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = MRSortInferenceAlternative
        fields = ("id", "category_id", "values")
        # This schema is only used nested, so `id` field is necessary
        # it won't be modifiable but will indicate which object to modify
        dump_only = ("values",)

    category_id = field_for(
        MRSortInferenceAlternative, "category_id", allow_none=True
    )
    values = fields.Function(
        lambda obj: {k.attribute_id: v for k, v in obj.values.items()}
    )


class MRSortInferenceCriterionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = MRSortCriterion
        fields = ("attribute_id", "mrsort_id", "maximize", "attribute")
        load_only = ("attribute_id", "mrsort_id")
        dump_only = ("attribute",)

    mrsort_id = fields.Integer()
    attribute_id = fields.Integer()
    attribute = fields.Nested(
        BaseDataAttributeSchema, only=("id", "name", "data")
    )


class MRSortInferenceSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = MRSort
        fields = ("criteria", "categories", "alternatives")
        dump_only = ("categories",)

    alternatives = fields.Nested(
        MRSortInferenceAlternativeSchema,
        many=True,
        attribute="inference_alternatives",
    )
    criteria = FieldMethod(
        fields.Nested(
            MRSortInferenceCriterionSchema(exclude=("mrsort_id",)), many=True
        ),
        "get_criteria",
        deserialize="load_criteria",
    )
    categories = fields.Nested(
        DiscreteCategorySchema, many=True, only=("name", "id")
    )

    def get_criteria(self, obj):
        schema = MRSortInferenceCriterionSchema(
            many=True, only=("attribute", "maximize")
        )
        data = schema.dump(obj.criteria)
        return data

    def load_criteria(self, value):
        new_criteria = []
        schema = MRSortInferenceCriterionSchema()
        for criterion_data in value:
            criterion_data["mrsort_id"] = self.instance.id
            instance = schema.get_instance(criterion_data)
            criterion_obj = schema.load(criterion_data, instance=instance)
            new_criteria.append(criterion_obj)
        return new_criteria



"""Remove project root data

Revision ID: 2657de160d2c
Revises: 89755dd43f58
Create Date: 2025-04-11 14:17:23.027086

"""

import sqlalchemy as sa
from alembic import op

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "2657de160d2c"
down_revision = "89755dd43f58"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table("project_root")


def downgrade():
    op.create_table(
        "project_root",
        sa.Column(
            "project_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "project_root_data_id",
            sa.INTEGER(),
            autoincrement=False,
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_project_root_project_id_project",
        ),
        sa.ForeignKeyConstraint(
            ["project_root_data_id"],
            ["project_data.id"],
            name="fk_project_root_project_root_data_id_project_data",
        ),
        sa.PrimaryKeyConstraint("project_id", name="pk_project_root"),
    )
    conn = op.get_bind()
    nb = conn.execute(sa.text("SELECT count(id) FROM project")).fetchone()[0]
    count = 0
    progress_bar(count, nb)
    for (project_id,) in conn.execute(sa.text("SELECT id FROM project")):
        pdata = conn.execute(
            sa.text(
                "SELECT id FROM project_data "
                f"WHERE project_id = {project_id} AND data_type = 'generator'"
            )
        ).fetchone()
        pdata = None if pdata is None else pdata[0]
        if pdata is None:
            # Create a blank DataGenerator named "Root"
            pdata = conn.execute(
                sa.text(
                    "INSERT INTO project_data "
                    "(data_type, project_id, name, description) "
                    f"VALUES ('generator', {project_id}, 'Root', '') "
                    "RETURNING id"
                )
            ).fetchone()[0]
            # Add default blank category_rules
            model = conn.execute(
                sa.text(
                    "INSERT INTO processing_model "
                    "(model_type, data_generator_id, name, cut_to_extent, "
                    "keep_overlap) "
                    f"VALUES ('categories_rule', {pdata}, '', true, 'max') "
                    "RETURNING id"
                )
            ).fetchone()[0]
            # Make it its active model
            conn.execute(
                sa.text(
                    f"UPDATE project_data SET active_model_id = {model} "
                    f"WHERE id = {pdata}"
                )
            )
        # Make project_data the root data of project
        conn.execute(
            sa.text(
                f"INSERT INTO project_root (project_id, project_root_data_id) "
                f"VALUES ({project_id}, {pdata})"
            )
        )
        count += 1
        progress_bar(count, nb)



import enum

from celery import states
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import backref

from . import db
from .mixins import ModelMixin


class ProjectTaskType(enum.Enum):
    process_project_data = 1
    update_stream = 2


class ProjectTaskModel(ModelMixin, db.Model):
    """Save the details of a celery task.

    Attributes:
        task_id (str): The id of the Celery task.
        started_at (datetime): When the task was started.
        finished_at (datetime): When the task was finished.
        state:
            The state of the task ('PENDING', 'STARTED', 'FAILED' or 'REVOKED')
        error_message (str): A description of the error if the task failed.
        type:
            The type of task
            ('process_project_data' or 'update_stream').
        params: The arguments used for the task.
        project_id: The id of the project in which this task is executed.
    """

    __tablename__ = "project_task"

    task_id = db.Column(db.String)
    started_at = db.Column(db.DateTime)
    finished_at = db.Column(db.DateTime)
    state = db.Column(db.String)
    error_message = db.Column(db.String)
    type = db.Column(db.Enum(ProjectTaskType))
    params = db.Column(JSONB)
    project_id = db.Column(
        db.Integer, db.ForeignKey("project.id"), nullable=False
    )

    # lazy loading for cascading the deletion of the project and zone
    # proposition
    project = db.relationship(
        "Project",
        backref=backref(
            "project_tasks", lazy="select", cascade="delete,delete-orphan"
        ),
    )

    def __init__(
        self, task_id, project=None, started_at=None, finished_at=None
    ):
        self.task_id = task_id
        self.project = project
        self.started_at = started_at
        self.finished_at = finished_at
        self.state = states.PENDING

    @classmethod
    def get_by_task_id(cls, task_id):
        """Returns the first task in the database with the given task id"""
        return cls.query.filter_by(task_id=task_id).first()



from geoalchemy2.shape import from_shape, to_shape
from marshmallow import fields
from shapely.geometry import box

from ..exceptions import InvalidValue
from ..models import Project, ProjectData, Template, User
from .base import BaseSchema, FieldMethod, FieldPluck
from .schemas import UserSchema


class ProjectSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Project
        exclude = ("_manager", "permissions", "project_tasks")
        dump_only = (
            "id",
            "is_template",
            "data_list",
            "created_at",
            "modified_at",
            "is_public",
            "manager",
        )

    name = fields.String()  # prevent the name field from being required
    manager = fields.Nested(UserSchema, only=("id", "username"))
    extent = FieldMethod(
        fields.List(fields.Float()),
        serialize="get_extent",
        deserialize="load_extent",
        allow_none=True,
        metadata={"minItems": 4, "maxItems": 4},
    )
    is_template = fields.Boolean(
        required=False
    )  # prevent complaint when field not set
    data_list = FieldPluck(ProjectData, "id", many=True)

    def get_extent(self, obj):
        """Returns the Geometry extent as a str "xmin,ymin,xmax,ymax" """
        if obj.extent is None:
            return None
        else:
            extent = to_shape(obj.extent)
            coords = [coord for coord in extent.bounds]
            return coords

    def load_extent(self, value):
        """Loads the extent "xmin,ymin,xmax,ymax" as a Geometry"""
        if value is not None:
            if (
                -180 <= value[0] <= 180
                and -90 <= value[1] <= 90
                and -180 <= value[2] <= 180
                and -90 <= value[3] <= 90
            ):
                extent = box(*value)
                return from_shape(extent, srid=4326)
            else:
                raise InvalidValue(
                    "Error in the extent value, should be within "
                    "[-180,-90,180,90]"
                )
        else:
            return None


class ProjectCreationSchema(ProjectSchema):
    name = fields.String(required=True, allow_none=False)


class ProjectPermissionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Project.Permission
        update_fields = ("user",)
        fields = ("user",)

    user = FieldPluck(User, "id")


class ProjectAccessSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Project
        update_fields = ("is_public", "permissions")
        fields = ("permissions", "is_public")

    permissions = fields.Nested(ProjectPermissionSchema, many=True)
    is_template = fields.Boolean(
        required=False
    )  # prevent complaint when field not set


class TemplateSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Template
        update_fields = ("name", "description")
        fields = (
            "id",
            "name",
            "description",
            "created_at",
            "manager",
        )
        dump_only = ("id", "created_at", "manager")

    manager = fields.Nested(UserSchema, only=("id", "username"))
    is_template = fields.Boolean(
        required=False
    )  # prevent complaint when field not set


class TemplatePermissionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Template.Permission
        update_fields = ("user",)
        fields = ("user",)

    user = FieldPluck(User, "id")


class TemplateAccessSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Template
        update_fields = ("is_public", "permissions")
        fields = ("permissions", "is_public")

    permissions = fields.Nested(TemplatePermissionSchema, many=True)
    is_template = fields.Boolean(
        required=False
    )  # prevent complaint when field not set



from flasgger import Swagger
from flask import Flask

swagger = Swagger()


def init_app(app: Flask):
    """Extract OpenAPI specification and create swagger"""
    from ..openapi.extract_specification import extract_openapi

    with app.app_context():
        spec = extract_openapi()
    swagger.template = spec.to_dict()
    swagger.init_app(app)



from getpass import getpass

from flask.cli import AppGroup

from ..models import PermissionAbility, User, UserPermission

user_cli = AppGroup("user", short_help="Enable manipulations of users")


@user_cli.command("create")
def create_user_command():
    """Create a user."""
    email = input("Email: ")
    if "@" not in email:
        raise ValueError("Invalid email, should contain the '@' character")
    _username = input("Username: ")
    username = email.split("@")[0] if len(_username) == 0 else _username
    password = getpass("Password: ")
    password_confirm = getpass("Confirm password: ")
    if password != password_confirm:
        raise ValueError("The password doesn't match")
    user = User(username=username, email=email, password=password)
    create_project = input("Allow to create projects? (y/n) [y]: ").lower()
    create_data = input("Allow to upload geo-data? (y/n) [y]: ").lower()
    if len(create_project) == 0 or create_data.lower()[0] == "y":
        user.permissions.append(
            UserPermission(ability=PermissionAbility.create_project)
        )
    if len(create_data) == 0 or create_data.lower()[0] == "y":
        user.permissions.append(
            UserPermission(ability=PermissionAbility.create_geo_data)
        )
    user.create()



from flask_restful import Api

from .auth import LoginAPI, RefreshAPI, UserAPI
from .geo_data import (
    DataAPI,
    DataAttributeAPI,
    DataDownloadAPI,
    DataListAPI,
    FeatureAPI,
    FeatureInputsAPI,
    FeatureOutputsAPI,
    GeoDataAccessAPI,
    GeoDataAPI,
    GeoDataListAPI,
    GeoDataUploadAPI,
    GeoDataUploadRasterAPI,
    GeoDataVectorTileAPI,
    GlobalDataAccessAPI,
    GlobalDataAPI,
    GlobalDataListAPI,
    StreamDataListAPI,
    StreamGeoDataAccessAPI,
    StreamGeoDataAPI,
    StreamGeoDataUpdateAPI,
    WFSGeoDataUploadAPI,
    WMSGeoDataUploadAPI,
)
from .permission import UserListAPI
from .project import (
    ProjectAPI,
    ProjectListAPI,
    ProjectPermissionsAPI,
    ProjectReplaceInputAttributesAPI,
    ProjectSharedDataListAPI,
    ProjectTemplateAPI,
    TemplateAPI,
    TemplateListAPI,
    TemplatePermissionsAPI,
)
from .project_data import (
    MRSortInferenceAPI,
    ProjectDataActiveModelAPI,
    ProjectDataAPI,
    ProjectDataInputSubgraphAPI,
    ProjectDataListAPI,
    ProjectDataModelAPI,
    ProjectDataModelChangeAPI,
    ProjectDataOutputSubgraphAPI,
    ProjectDataReplaceInputAttributesAPI,
)
from .share import (
    DataShareAPI,
    ProjectSharesAPI,
    ShareBaseDataAPI,
    ShareDataListAPI,
)
from .tasks import ProjectTaskListAPI, TaskAPI

api = Api()

api.add_resource(DataListAPI, "/data")
api.add_resource(DataAPI, "/data/<int:data_id>")

api.add_resource(GlobalDataListAPI, "/global-data")
api.add_resource(GlobalDataAPI, "/global-data/<int:data_id>")
api.add_resource(GlobalDataAccessAPI, "/global-data/<int:data_id>/permissions")

api.add_resource(GeoDataListAPI, "/geo-data")
api.add_resource(
    GeoDataVectorTileAPI,
    "/geo-data/<int:geo_data_id>/tiles/<int:z>/<int:x>/<int:y>",
)
api.add_resource(GeoDataAPI, "/geo-data/<int:geo_data_id>")
api.add_resource(GeoDataAccessAPI, "/geo-data/<int:geo_data_id>/permissions")
api.add_resource(DataDownloadAPI, "/data/<int:data_id>/export")
api.add_resource(GeoDataUploadAPI, "/geo-data/upload")
api.add_resource(GeoDataUploadRasterAPI, "/geo-data/upload-raster")
api.add_resource(StreamDataListAPI, "/stream-geo-data")
api.add_resource(StreamGeoDataAPI, "/stream-geo-data/<int:stream_geo_data_id>")
api.add_resource(
    StreamGeoDataAccessAPI,
    "/stream-geo-data/<int:stream_geo_data_id>/permissions",
)
api.add_resource(
    StreamGeoDataUpdateAPI,
    "/stream-geo-data/<int:stream_geo_data_id>/update-data",
)
api.add_resource(WFSGeoDataUploadAPI, "/stream-geo-data/upload-wfs")
api.add_resource(WMSGeoDataUploadAPI, "/stream-geo-data/upload-wms")
api.add_resource(DataAttributeAPI, "/attribute/<int:attribute_id>")
api.add_resource(FeatureAPI, "/feature/<int:feature_id>")
api.add_resource(
    FeatureInputsAPI, "/feature/<int:feature_id>/input-data/<int:data_id>"
)
api.add_resource(
    FeatureOutputsAPI, "/feature/<int:feature_id>/output-data/<int:data_id>"
)

api.add_resource(ProjectListAPI, "/projects")
api.add_resource(ProjectAPI, "/project/<int:project_id>")
api.add_resource(
    ProjectPermissionsAPI, "/project/<int:project_id>/permissions"
)
api.add_resource(ProjectSharedDataListAPI, "/project/<int:project_id>/data")
api.add_resource(ProjectTaskListAPI, "/project/<int:project_id>/tasks")
api.add_resource(ProjectSharesAPI, "/project/<int:project_id>/shared")
api.add_resource(ProjectTemplateAPI, "/project/<int:project_id>/template")
api.add_resource(
    ProjectReplaceInputAttributesAPI,
    "/project/<int:project_id>/replace-attributes",
)
api.add_resource(TemplateListAPI, "/templates")
api.add_resource(TemplateAPI, "/template/<int:template_id>")
api.add_resource(
    TemplatePermissionsAPI, "/template/<int:template_id>/permissions"
)

api.add_resource(TaskAPI, "/task/<string:task_id>")

api.add_resource(ProjectDataListAPI, "/project-data")
api.add_resource(ProjectDataAPI, "/project-data/<int:data_id>")
api.add_resource(
    ProjectDataActiveModelAPI, "/project-data/<int:data_id>/model"
)
api.add_resource(
    ProjectDataModelAPI, "/project-data/<int:data_id>/model/<int:model_id>"
)
api.add_resource(
    ProjectDataModelChangeAPI, "/project-data/<int:data_id>/model/change"
)
api.add_resource(
    MRSortInferenceAPI, "/project-data/<int:data_id>/mr-sort-inference"
)
api.add_resource(
    ProjectDataInputSubgraphAPI, "/project-data/<int:data_id>/input-graph"
)
api.add_resource(
    ProjectDataOutputSubgraphAPI, "/project-data/<int:data_id>/output-graph"
)
api.add_resource(
    ProjectDataReplaceInputAttributesAPI,
    "/project-data/<int:data_id>/replace-attributes",
)
api.add_resource(ShareBaseDataAPI, "/data/<int:data_id>/share")
api.add_resource(ShareDataListAPI, "/shares")
api.add_resource(DataShareAPI, "/shares/<string:uid>")

api.add_resource(UserListAPI, "/users")
api.add_resource(LoginAPI, "/login")
api.add_resource(RefreshAPI, "/refresh")
api.add_resource(UserAPI, "/user")



from flask import jsonify, request
from flask_restful import Resource

from ..exceptions import RequestError
from ..models import User
from ..schemas import (
    UserAccessTokenResponseSchema,
    UserSchema,
    UserTokenResponseSchema,
)
from ..security import jwt_handler
from ..services import auth_service
from .utils import dump_data, with_response


class LoginAPI(Resource):
    def create_user_token_response(self, user):
        """
        Create a new authentication and refresh token for the user.
        """
        token = user.create_jwt()
        refresh = user.create_refresh_token()
        data = {
            "access_token": token,
            "refresh_token": refresh,
            "token_type": jwt_handler.header_prefix,
        }
        return data

    @auth_service.basic_authenticated
    @with_response(
        status=200,
        schema=UserTokenResponseSchema,
        description="User access and refresh tokens",
    )
    def get(self, user: User):
        """
        Log the user in and return an authentication and a refresh token.

        :reqheader Authorization: `Basic authentication <https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication#Basic_authentication_scheme>`_, ``user:password`` encoded in base64.
        :resjson str access_token: Access token.
        :resjson str refresh_token:
            Token to use for creating a new authentication token.
        :resjson str token_type: Type of the token

        .. :quickref: Authentication; Log the user in"""  # noqa: E501
        data = self.create_user_token_response(user)
        user.update()
        return jsonify(**data)


class RefreshAPI(Resource):
    def create_new_jwt(self, user):
        """
        Create a new authentication token for the user.
        """
        token = user.create_jwt()
        data = {"access_token": token, "token_type": jwt_handler.header_prefix}
        return data

    @auth_service.refresh_token_required
    @with_response(
        status=200,
        schema=UserAccessTokenResponseSchema,
        description="Refreshed user access token",
    )
    def get(self, user: User):
        """
        Create a new authentication token from the refresh token.

        :reqheader Authorization: The refresh token received when login in.
        :resjson str access_token: Access token.
        :resjson str token_type: Type of the token

        .. :quickref: Authentication; Create a new authentication token
        """
        data = self.create_new_jwt(user)
        return jsonify(**data)


class UserAPI(Resource):
    @auth_service.token_required
    @with_response(
        status=200, schema=UserSchema, description="Current user details"
    )
    def get(self):
        """
        Return the details of the current user.

        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).

        .. :quickref: User; Get the details of the current user
        """
        user = auth_service.check_jwt_authentication(request)
        if user is None:
            raise RequestError("No user connected", 401)
        else:
            return dump_data(UserSchema(), user=user)



from flask import current_app as app
from marshmallow import (
    Schema,
    fields,
    pre_dump,
    pre_load,
    validate,
    validates_schema,
)
from marshmallow.exceptions import ValidationError

from ..models import DataAttribute
from ..models.processing_models import (
    AttributeType,
    ContinuousRule,
    DefaultValue,
    DiscreteCategory,
    DiscreteRules,
    DiscreteRulesCategory,
    DissolveAdjacentModel,
    GeoBuffer,
    KeepOverlap,
    MergeOverlapModel,
    ModelType,
    MRSort,
    MRSortCriterion,
    PrefDefaultValues,
    ProcessingModel,
    WeightedSum,
    WeightedSumOperand,
    ZonePropositionGenerator,
)
from .base import (
    BaseSchema,
    FieldMethod,
    FieldPluck,
    OneOfSchema,
    OneOfSchemaWithType,
    constant_enum,
)
from .geo_data import BaseDataAttributeSchema


class PrefModelBaseSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = ProcessingModel
        dump_only = ("id", "model_type", "data_generator", "used_attributes")

    model_type = fields.Enum(ModelType, required=True)
    data_generator = fields.Pluck("DataGeneratorSchema", "id")
    used_attributes = FieldMethod(
        fields.Nested(
            BaseDataAttributeSchema,
            only=("id", "name", "data", "type"),
            many=True,
        ),
        "get_used_attributes",
    )

    def get_used_attributes(self, obj):
        return [
            {
                "id": attribute.id,
                "name": attribute.name,
                "data": {"id": attribute.data_id, "name": attribute.data.name},
                "type": attribute.type.name,
            }
            for attribute in obj.get_used_input_attributes()
        ]


class PrefModelBaseCreationSchema(PrefModelBaseSchema):
    pass


class MergeOverlapModelSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = MergeOverlapModel
        fields = (
            "id",
            "model_type",
            "data_generator",
            "cut_to_extent",
            "keep_overlap",
            "name",
            "used_attributes",
        )

    keep_overlap = fields.Enum(KeepOverlap)
    model_type = fields.Enum(
        constant_enum(ModelType.merge_overlap), required=True
    )


class MergeOverlapModelCreationSchema(MergeOverlapModelSchema):
    pass


class DissolveAdjacentModelSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = DissolveAdjacentModel

    model_type = fields.Enum(
        constant_enum(ModelType.dissolve_adjacent), required=True
    )


class DissolveAdjacentModelCreationSchema(DissolveAdjacentModelSchema):
    pass


class GeoBufferSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = GeoBuffer
        fields = (
            "id",
            "model_type",
            "data_generator",
            "cut_to_extent",
            "radius",
            "name",
            "used_attributes",
        )

    model_type = fields.Enum(
        constant_enum(ModelType.geo_buffer), required=True
    )


class GeoBufferCreationSchema(GeoBufferSchema):
    pass


class DiscreteCategorySchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = DiscreteCategory
        fields = ("id", "name", "position")
        load_only = ("position",)


class DiscreteRulesCategorySchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = DiscreteRulesCategory
        fields = ("id", "name", "position", "rules")


class DiscreteRulesSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = DiscreteRules
        fields = (
            "id",
            "model_type",
            "data_generator",
            "cut_to_extent",
            "categories",
            "name",
            "used_attributes",
        )

    model_type = fields.Enum(
        constant_enum(ModelType.categories_rule), required=True
    )
    categories = fields.Nested(
        DiscreteRulesCategorySchema, many=True, only=("name", "rules")
    )


class DiscreteRulesCreationSchema(DiscreteRulesSchema):
    pass


class MRSortCriterionSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = MRSortCriterion
        fields = (
            "attribute_id",
            "mrsort_id",
            "attribute",
            "profiles",
            "weight",
            "maximize",
        )
        load_only = ("attribute_id", "mrsort_id")
        dump_only = ("attribute",)

    attribute_id = fields.Integer()
    mrsort_id = fields.Integer()
    attribute = fields.Nested(
        BaseDataAttributeSchema, only=("id", "name", "data")
    )


class MRSortSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = MRSort
        fields = (
            "id",
            "model_type",
            "data_generator",
            "cut_to_extent",
            "categories",
            "criteria",
            "majority_threshold",
            "name",
            "used_attributes",
        )

    model_type = fields.Enum(constant_enum(ModelType.mrsort), required=True)
    categories = FieldMethod(
        fields.Nested(DiscreteCategorySchema, many=True),
        serialize="get_categories",
        deserialize="load_categories",
    )
    criteria = FieldMethod(
        fields.Nested(
            MRSortCriterionSchema(exclude=("mrsort_id",)), many=True
        ),
        "get_criteria",
        deserialize="load_criteria",
    )

    def get_categories(self, obj):
        schema = DiscreteCategorySchema(many=True)
        data = schema.dump(obj.categories)
        return data

    def load_categories(self, value):
        """Load the categories from a dictionary.

        If there is already an existing category with the same name,
        reuse the same object instance.
        """
        new_categories = []
        schema = DiscreteCategorySchema()
        # Check if the MR-Sort instance exists
        if self.instance is not None:
            categories = {c.name: c for c in self.instance.categories}
        else:
            categories = {}
        # Create the categories instances
        for category_data in value:
            name = category_data.get("name")
            # Get the category with the same name if it exists
            instance = categories.get(name, None)
            category_obj = schema.load(category_data, instance=instance)
            new_categories.append(category_obj)
        # Set the position used for the ordering in the database
        for position, category in enumerate(new_categories):
            category.position = position
        return new_categories

    def get_criteria(self, obj):
        schema = MRSortCriterionSchema(
            many=True, only=("attribute", "profiles", "weight", "maximize")
        )
        data = schema.dump(obj.criteria)
        return data

    def load_criteria(self, value):
        new_criteria = []
        schema = MRSortCriterionSchema()
        for criterion_data in value:
            criterion_data["mrsort_id"] = self.instance.id
            instance = schema.get_instance(criterion_data)
            criterion_obj = schema.load(criterion_data, instance=instance)
            new_criteria.append(criterion_obj)
        return new_criteria

    @pre_dump
    def init_criteria(self, mrsort, **kwargs):
        """Load the criteria of the MR-Sort data"""
        app.logger.info("Update the MR-Sort criteria")
        mrsort.init_criteria()
        mrsort.update()
        return mrsort


class MRSortCreationSchema(MRSortSchema):
    pass


class DefaultValueSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = DefaultValue
        fields = ("attribute", "value")

    attribute = FieldPluck(DataAttribute, "id")


class PrefDefaultValuesAttributeSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    attribute = fields.String()
    type = fields.Enum(AttributeType, required=True)


class PrefDefaultValuesBaseSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema):
        model = PrefDefaultValues

    # As it has no deserialize method we don't need to add it to the dump_only
    # fields
    attributes_list = FieldMethod(
        fields.Nested(PrefDefaultValuesAttributeSchema, many=True),
        "get_attributes_list",
    )
    default_values = fields.Nested(
        DefaultValueSchema, many=True, only=("attribute", "value")
    )

    def get_attributes_list(self, obj):
        attrs = []
        for input_data in obj.data_generator.input_data:
            if input_data.data is None:
                continue
            for attr in input_data.data.attributes:
                attrs.append(
                    {
                        "id": attr.id,
                        "name": input_data.name,
                        "attribute": attr.name,
                        "type": attr.type.name,
                    }
                )
        return attrs


class WeightedSumOperandSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = WeightedSumOperand

    attribute = FieldPluck(DataAttribute, "id")


class WeightedSumSchema(PrefDefaultValuesBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = WeightedSum

    model_type = fields.Enum(
        constant_enum(ModelType.weighted_sum), required=True
    )
    operands = fields.Nested(
        WeightedSumOperandSchema, many=True, only=("attribute", "weight")
    )


class WeightedSumCreationSchema(WeightedSumSchema):
    pass


class ContinuousRuleSchema(PrefDefaultValuesBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = ContinuousRule

    model_type = fields.Enum(
        constant_enum(ModelType.continuous_rule), required=True
    )


class ContinuousRuleCreationSchema(ContinuousRuleSchema):
    pass


class ZonePropositionGeneratorSchema(PrefModelBaseSchema):
    class Meta(PrefModelBaseSchema.Meta):
        model = ZonePropositionGenerator

    geo_size = fields.Float(required=False, allow_none=False)
    model_type = fields.Enum(
        constant_enum(ModelType.zone_proposition), required=True
    )
    size = fields.Integer(
        validate=validate.Range(1, 1000), required=False, allow_none=False
    )
    mutation = fields.Float(
        validate=validate.Range(0, 1), required=False, allow_none=False
    )
    children = fields.Integer(
        validate=validate.Range(0, None), required=False, allow_none=False
    )
    filter_clusters = fields.Boolean(required=False, allow_none=False)

    @validates_schema
    def validate_children(self, data, **kwargs):
        if self.get_future_field(data, "children") > self.get_future_field(
            data, "size"
        ):
            raise ValidationError(
                "Cannot create more children than the parent population"
            )

    @validates_schema
    def validate_end_condition(self, data, **kwargs):
        if (
            self.get_future_field(data, "duration") is None
            and self.get_future_field(data, "iterations") is None
        ):
            raise ValidationError("Need at least one exit condition")


class ZonePropositionGeneratorCreationSchema(ZonePropositionGeneratorSchema):
    geo_size = fields.Float(required=False, allow_none=False, missing=0)
    size = fields.Integer(
        validate=validate.Range(1, 1000), missing=80, allow_none=False
    )
    mutation = fields.Float(
        validate=validate.Range(0, 1), missing=0.02, allow_none=False
    )
    children = fields.Integer(
        validate=validate.Range(0, None), missing=60, allow_none=False
    )
    filter_clusters = fields.Boolean(missing=True, allow_none=False)

    @pre_load
    def default_end_condition(self, data, **kwargs):
        if "duration" in data or "iterations" in data:
            return data
        data["iterations"] = 1
        return data


class ProcessingModelSchema(OneOfSchema, PrefModelBaseSchema):
    type_field = "model_type"
    type_schemas = {
        ModelType.categories_rule.name: DiscreteRulesSchema,
        ModelType.continuous_rule.name: ContinuousRuleSchema,
        ModelType.geo_buffer.name: GeoBufferSchema,
        ModelType.mrsort.name: MRSortSchema,
        ModelType.weighted_sum.name: WeightedSumSchema,
        ModelType.merge_overlap.name: MergeOverlapModelSchema,
        ModelType.dissolve_adjacent.name: DissolveAdjacentModelSchema,
        ModelType.zone_proposition.name: ZonePropositionGeneratorSchema,
    }

    def get_obj_type(self, obj):
        return obj.model_type.name


class ProcessingModelCreationSchema(
    OneOfSchemaWithType, PrefModelBaseCreationSchema
):
    type_field = "model_type"
    type_schemas = {
        ModelType.categories_rule.name: DiscreteRulesCreationSchema,
        ModelType.continuous_rule.name: ContinuousRuleCreationSchema,
        ModelType.geo_buffer.name: GeoBufferCreationSchema,
        ModelType.mrsort.name: MRSortCreationSchema,
        ModelType.weighted_sum.name: WeightedSumCreationSchema,
        ModelType.merge_overlap.name: MergeOverlapModelCreationSchema,
        ModelType.dissolve_adjacent.name: DissolveAdjacentModelCreationSchema,
        ModelType.zone_proposition.name: ZonePropositionGeneratorCreationSchema,  # noqa: E501
    }

    def get_obj_type(self, obj):
        return obj.model_type.name



import importlib.resources as pkg_resources

import click
from alembic import command
from alembic.migration import MigrationContext
from flask import Flask
from sqlalchemy.sql import text

from .. import sql
from ..models import db


def register_commands(app: Flask):
    db_cli_group = app.cli.commands["db"]

    @db_cli_group.command()
    @click.option(
        "-d",
        "--directory",
        default=None,
        help=('Migration script directory (default is "migrations")'),
    )
    @click.option(
        "--tag",
        default=None,
        help=(
            'Arbitrary "tag" name - can be used by custom env.py ' "scripts"
        ),
    )
    @click.option(
        "-f",
        "--force",
        default=False,
        is_flag=True,
        help="Force database initialization (even if already initialized)",
    )
    @click.option(
        "-x",
        "--x-arg",
        multiple=True,
        help="Additional arguments consumed by custom env.py scripts",
    )
    def create(directory, tag, force, x_arg):
        """Creates all the SQLAlchemy tables and add SQL functions."""

        schema = app.config.get("DB_SCHEMA", "public")
        config = app.extensions["migrate"].migrate.get_config(
            directory, x_arg=x_arg
        )

        engine = db.get_engine()
        with engine.connect() as conn:
            conn.execute(text(f"CREATE SCHEMA IF NOT EXISTS {schema}"))
            conn.execute(text("COMMIT"))  # commit immediately
            migration_ctx = MigrationContext.configure(conn)
            current = migration_ctx.get_current_revision()

        if current is None or force:
            db.create_all()
            with (
                pkg_resources.files(sql)
                .joinpath("dissolve_adjacent.sql")
                .open("r") as f
            ):
                db.session.execute(text(f.read()))
            with (
                pkg_resources.files(sql)
                .joinpath("tile_bbox.sql")
                .open("r") as f
            ):
                db.session.execute(text(f.read()))
            with (
                pkg_resources.files(sql)
                .joinpath("trigger_feature_make_valid.sql")
                .open("r") as f
            ):
                db.session.execute(text(f.read()))
            db.session.commit()
            command.stamp(config, revision="head", tag=tag)
        else:
            print(
                "Database already exists. You may force its creation "
                "with '--force' option (may break database).\n"
                "If wanting to upgrade it, use 'flask db upgrade' command "
                "instead."
            )

    @db_cli_group.command()
    @click.option(
        "-d",
        "--directory",
        default=None,
        help=('Migration script directory (default is "migrations")'),
    )
    @click.option(
        "-x",
        "--x-arg",
        multiple=True,
        help="Additional arguments consumed by custom env.py scripts",
    )
    def drop(directory, x_arg):
        """Drops all the SQLAlchemy tables."""

        config = app.extensions["migrate"].migrate.get_config(
            directory, x_arg=x_arg
        )

        if input("Your data will be lost. Continue? (Y/y): ")[0] in ["Y", "y"]:
            db.drop_all()
            command.stamp(config, revision=None)

    @db_cli_group.command("migrate-schema")
    @click.argument("new_schema")
    def migrate_schema(new_schema: str):
        """
        Move ORM tables (and alembic_version) to NEW_SCHEMA.

        Example:
            flask migrate-schema public deseasion
        """

        # Normalize: treat "public" as default schema
        def normalize_schema(schema: str | None) -> str:
            return schema if schema is not None else "public"

        engine = db.get_engine()
        with engine.begin() as conn:
            # 1. Ensure new schema exists (unless it's "public")
            if new_schema != "public":
                conn.execute(text(f"CREATE SCHEMA IF NOT EXISTS {new_schema}"))
            old_schema = normalize_schema(db.metadata.schema)

            if old_schema == new_schema:
                click.echo(
                    f"No migration necessary: schema is already {new_schema}"
                )
                return

            # 2. Move ORM tables
            for table_name in db.metadata.tables:
                fq_table = (
                    f"{old_schema}.{table_name}"
                    if old_schema != "public"
                    else table_name
                )
                fq_table_quoted = (
                    f'"{old_schema}"."{table_name}"'
                    if old_schema != "public"
                    else f'"{table_name}"'
                )
                click.echo(f"Moving {fq_table} → {new_schema}.{table_name}")
                conn.execute(
                    text(
                        f"ALTER TABLE {fq_table_quoted} "
                        f"SET SCHEMA {new_schema}"
                    )
                )

            # 3. Move alembic_version table
            version_table_old = (
                f"{db.metadata.schema}.alembic_version"
                if db.metadata.schema is not None
                and db.metadata.schema != "public"
                else "alembic_version"
            )
            res = conn.execute(
                text("SELECT to_regclass(:tbl)"),
                {"tbl": version_table_old},
            ).scalar()

            if res:
                click.echo(
                    f"Moving alembic_version → {new_schema}.alembic_version"
                )
                conn.execute(
                    text(
                        f"ALTER TABLE {version_table_old} "
                        f"SET SCHEMA {new_schema}"
                    )
                )

            click.echo("✅ Schema migration completed.")
            click.echo(
                "⚠️ Make sure to add the instance config.py file "
                f'following line: DB_SCHEMA = "{new_schema}"'
            )



import re
from dataclasses import dataclass

import bs4
import geopandas as gpd
from fiona import BytesCollection
from fiona.errors import DriverError

from ..exceptions import (
    CapabilitiesXMLParsingError,
    ExternalRequestError,
    UploadError,
)
from .requests import get


@dataclass
class WFSFeatureType:
    """This dataclass is meant for holding information parsed from
    a <FeatureType> XML tag in a WFS GetCapabilities query response.

    .. todo::
        Add other srs options from server, and use them to make the
        projection server-side if possible. Must be able to compare srs
        on multiple formats for that...
    """

    name: str
    title: str
    description: str
    keywords: list[str]
    srs: list[str]

    @property
    def default_srs(self) -> str:
        """Default SRS projection (first one from :attr:`srs` list)"""
        return self.srs[0]


@dataclass
class WFSCapabilities:
    """This dataclass holds information parsed from a WFS GetCapabilities query
    XML response.
    """

    url: str
    version: str
    feature_types: dict[str, WFSFeatureType]


def version2int(version: str) -> tuple[int, ...]:
    """Convert version string to int tuple.

    :param version:
    :return: version int tuple in same order
    """
    return tuple(int(v) for v in version.split("."))


def convert_raw_crs(raw_crs: str) -> str:
    """Convert a crs/srs XML tag content into a usable srs string.

    Formats converted:

    * http://www.opengis.net/def/crs/epsg/0/<EPSG code>
    * http://www.opengis.net/gml/srs/epsg.xml#<EPSG code>

    Any other formats than the first 2 are passed as is, so all WFS standard
    srs formats are handled.

    :param raw_crs: crs/srs XML tag content
    :return: srs/csr usable for projections
    """
    if raw_crs.startswith("http") and "epsg.xml" in raw_crs:
        return f"EPSG:{raw_crs.split('#')[-1]}"
    if raw_crs.startswith("http") and "epsg" in raw_crs:
        return f"EPSG:{raw_crs.split('/')[-1]}"
    return raw_crs


def parse_feature_type(feature_type: bs4.Tag, version: str) -> WFSFeatureType:
    """Parse a XML <FeatureType> tag.

    :param feature_type: XML <FeatureType> tag
    :param version: WFS version
    :return: information describing the feature type

    .. todo:: Add other srs/crs options to srs
    """
    version_int = version2int(version)
    name = "".join(feature_type.find("Name").contents)
    title = "".join(feature_type.find("Title").contents)
    description = "\n".join(feature_type.find("Abstract").contents)
    keywords = []
    keyword_tag = feature_type.find("Keywords")
    for key in keyword_tag.contents:
        match key:
            case str():
                if len(re.findall(r"\S", key)) == 0:
                    continue
                keywords.append(key.strip())
            case bs4.Tag(name="Keyword"):
                keywords.append("".join(key.contents).strip())
    srs = []
    if version_int < (1, 1):
        srs.append(
            convert_raw_crs("".join(feature_type.find("SRS").contents).strip())
        )
    elif version_int >= (2, 0):
        srs.append(
            convert_raw_crs(
                "".join(feature_type.find("DefaultCRS").contents).strip()
            )
        )
    else:
        srs.append(
            convert_raw_crs(
                "".join(feature_type.find("DefaultSRS").contents).strip()
            )
        )

    return WFSFeatureType(name, title, description, keywords, srs)


def parse_capabilities(
    capabilities: bs4.Tag,
) -> tuple[str, dict[str, WFSFeatureType]]:
    """Parse a <WFS_Capabilities> XML tag for information.

    :param capabilities: XML tag
    :raises CapabilitiesXMLParsingError:
        if XML parsing of `capabilities` failed
    :return: WFS server capabilities
    """
    if capabilities is None:
        raise CapabilitiesXMLParsingError("cannot find tag 'WFS_Capabilities'")
    if not isinstance(capabilities, bs4.element.Tag):
        raise CapabilitiesXMLParsingError(
            f"Malformed GetCapabilities XML: 'WFS_Capabilities' should be a "
            f"tag instead of a {type(capabilities)}"
        )
    if "version" not in capabilities.attrs:
        raise CapabilitiesXMLParsingError(
            "Malformed GetCapabilities XML: cannot find 'version'"
        )
    version = capabilities.attrs.get("version")
    layers = {}
    for feature_tag in capabilities.find_all("FeatureType"):
        try:
            layer = parse_feature_type(feature_tag, version)
            layers[layer.name] = layer
        except (AttributeError, IndexError) as e:
            raise CapabilitiesXMLParsingError(str(e))
    return version, layers


def get_capabilities(url: str, version: str = None) -> WFSCapabilities:
    """Perform WFS GetCapabilities request and parse informations.

    :param url: WFS server URL
    :param version: WFS version number, negotiated with server if not set
    :raises CapabilitiesXMLParsingError:
        if XML parsing of `capabilities` failed
    :raises ExternalRequestError:
        if GetCapabilities request to WFS server failed
    :return: WFS server capabilities
    """
    params = {
        "service": "WFS",
        "request": "GetCapabilities",
    }
    if version is not None:
        params["version"] = version
    response = get(url, params=params, request_type="GetCapabilities")
    soup = bs4.BeautifulSoup(response.content, features="xml")
    return WFSCapabilities(
        url, *parse_capabilities(soup.find("WFS_Capabilities"))
    )


def get_feature(
    url: str, feature_type: str, version: str = None, srs: str = None, **kwargs
) -> gpd.GeoDataFrame:
    """Perform WFS GetFeature request and return geographic data.

    :param url: WFS server URL
    :param feature_type: feature type to retrieve
    :param version:
        preferred WFS version number (actual is negotiated with server)
    :param srs: coordinate system to use, use feature type default if not set
    :raises CapabilitiesXMLParsingError:
        if XML parsing of GetCapabilities response failed
    :raises ExternalRequestError:
        if GetCapabilities or GetFeature request to WFS server failed
    :raises UploadError:
        if the GetFeature response could not be loaded into geodata
    :return: geographic data
    """
    capabilities = get_capabilities(url, version)
    layer_data = capabilities.feature_types[feature_type]
    version = capabilities.version
    params = {
        "service": "WFS",
        "request": "GetFeature",
        "version": version,
    }
    version_int = version2int(version)
    if version_int >= (2, 0):
        params["typeNames"] = [feature_type]
    else:
        params["typeName"] = feature_type
    if srs is not None:
        params["srsName"] = srs
    params = {**params, **kwargs}
    response = get(url, params=params, request_type="GetFeature")
    try:
        with BytesCollection(response.content) as features:
            df = gpd.GeoDataFrame.from_features(
                features, crs=(srs or layer_data.default_srs)
            )
        return df
    except DriverError:
        try:
            if "xml" not in response.headers.get("content-type", "").lower():
                raise UploadError(
                    f"Cannot load geodata from URL: {response.url}"
                )
            soup = bs4.BeautifulSoup(response.content, features="xml")
            if version_int >= (1, 1):
                etag = soup.find("Exception")
                sub_etag = etag.find("ExceptionText")
                srv_msg = f"{etag.attrs['exceptionCode']}: " + "".join(
                    sub_etag.contents
                ).replace("\n", "")
            else:
                etag = soup.find("ServiceException")
                srv_msg = f"{etag.attrs['code']}: " + "".join(
                    etag.contents
                ).replace("\n", "")
        except (AttributeError, IndexError):
            raise UploadError(f"Cannot load geodata from URL: {response.url}")
        raise ExternalRequestError(
            response.url,
            f"WFS server cannot return geodata. {srv_msg}",
            request_type="GetFeature",
        )



from flask import jsonify, request, send_file
from flask_restful import Resource, reqparse

from ..exceptions import RequestError
from ..models import DataShare
from ..schemas import (
    DataShareSchema,
    MessageSchema,
    ProjectSharesGetResponseSchema,
    ShareBaseDataPostResponseSchema,
)
from ..services import data_service, project_service
from ..services.auth_service import check_jwt_authentication, token_required
from ..services.geo_data_loading_service import (
    get_geo_data_download,
    get_temporary_directory,
)
from .utils import (
    dump_data,
    get_json_content,
    with_query_arg,
    with_request_body,
    with_response,
)


class ShareDataListAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectSharesGetResponseSchema,
        description="List all data shares",
    )
    def get(self):
        """
        Get the list of data shares.

        :status 403: The user is not logged in.

        .. :quickref: Shares; Get list of data shares
        """
        user = check_jwt_authentication(request)
        if user is None:
            raise PermissionError("No user logged in")
        shares = DataShare.query.all()
        shared = list(
            filter(
                lambda s: not s.is_expired()
                and data_service.has_permission(s.data, user),
                shares,
            )
        )
        schema = DataShareSchema(many=True)
        return dump_data(schema, shared=shared)


class ProjectSharesAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectSharesGetResponseSchema,
        description="List of data shares of the project",
    )
    def get(self, project_id):
        """
        Get the list of data shares of a project.

        :param int project_id: The id of the project.
        :status 403: The user is not allowed to access the project.
        :status 404: The project does not exist.

        .. :quickref: Project; Get list of data shares
        """
        shared = project_service.get_authorized_shares(project_id)
        schema = DataShareSchema(many=True)
        return dump_data(schema, shared=shared)


class ShareBaseDataAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ShareBaseDataPostResponseSchema,
        description="New data share details",
    )
    @with_request_body(
        schema=DataShareSchema, description="New data share details to create"
    )
    def post(self, data_id):
        """
        Create a new data share.

        :param int data_id: The id of the data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project data.
        :status 404: The project data does not exist.

        .. :quickref: Data; Share data
        """
        data = data_service.get_if_authorized(data_id)
        content = get_json_content()
        schema = DataShareSchema()
        share = schema.load(content)
        share.data = data
        share.create()
        return dump_data(schema, share=share)

    @token_required
    @with_response(
        200,
        schema=ProjectSharesGetResponseSchema,
        description="List of shares of the data",
    )
    def get(self, data_id):
        """
        Get the list of shares of a data.

        :param int data_id: The id of the data.
        :status 403: The user is not allowed to access the data.
        :status 404: The data does not exist.

        .. :quickref: Data; Get list of data shares
        """
        data = data_service.get_if_authorized(data_id)
        shared = []
        for share in data.shares:
            if not share.is_expired():
                shared.append(share)
        schema = DataShareSchema(many=True)
        return dump_data(schema, shared=shared)


class DataShareAPI(Resource):
    @with_query_arg(
        "outputFormat",
        str,
        required=False,
        description="Desired output format (leave blank for zip shapefile)",
    )
    @with_response(
        status=200,
        description="Geodata in specified file format",
        content={
            "application/octet-stream": {"type": "string", "format": "binary"}
        },
    )
    def get(self, uid):
        """Return the geometries of a shared data.

        :param str uid: The uid of a data share.
        :status 404: The share does not exist at this time.
        :status 415: Output format is not supported.
        :status 500:
            There was an error with the geometry while creating the file or
            data type is currently not supported for download.

        .. :quickref: Shares; Download a shared geodata
        """
        parser = reqparse.RequestParser()
        parser.add_argument("outputFormat", type=str, location="args")
        args = parser.parse_args()

        share = DataShare.get_by_uid(uid)
        if share is None or share.is_expired():
            raise RequestError("Share not existing", 404)
        data = share.data
        with get_temporary_directory() as temp_dir:
            filename, file_object = get_geo_data_download(
                data, temp_dir, args.get("outputFormat")
            )
            return send_file(
                file_object, as_attachment=True, download_name=filename
            )

    @token_required
    @with_response(
        200,
        schema=MessageSchema,
        description="Data share deleted successfully",
    )
    def delete(self, uid):
        """Delete a data share.

        :param str uid: The uid of a data share.
        :status 403: The user cannot delete share on this project.
        :status 404: The share or its project data does not exist.

        .. :quickref: Shares; Delete a data share
        """
        share = DataShare.get_by_uid(uid)
        if share is None or share.is_expired():
            raise RequestError("Share not existing", 404)
        data_service.get_if_authorized(share.data_id)
        share.expired = True
        share.update()
        return jsonify(message="share {} set to expired".format(uid))



"""add project template permissions

Revision ID: c6e3da96b784
Revises: 2657de160d2c
Create Date: 2025-05-27 16:27:22.815170

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c6e3da96b784"
down_revision = "2657de160d2c"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        sa.text(
            "INSERT INTO project_permission (object_id, user_id) "
            "SELECT id, manager_id FROM project WHERE is_template=true"
        )
    )


def downgrade():
    op.execute(
        sa.text(
            "DELETE FROM project_permission WHERE object_id IN ("
            "SELECT id FROM project WHERE is_template=true"
            ")"
        )
    )



import enum

from marshmallow import fields
from marshmallow.utils import is_collection
from marshmallow_oneofschema import OneOfSchema as _OneOfSchema
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema

from ..models import db


def constant_enum(value) -> type[enum.Enum]:
    """This creates an enum with a single value.

    Its use should be limited to schema polymorphism so a type discriminator
    which is an enumeration of all possible types (children types) is set
    as only the child type in the child.

    :param value: either a single enum item or any object
    :return: enumeration class with only `value` as item

    .. note::
        might override a String field as well if the parent class type
        discriminator field was left open for future appendings, this is
        generally how schema polymorphism is implemented in OpenAPI 3.0+.
        In this case, the closed enumeration must be done in the oneOf classes.
    """
    match value:
        case enum.Enum():
            return enum.Enum(f"enum{value}", {value.name: value.value})
        case _:
            return enum.Enum(f"enum{value}", {value: value})


class FieldMethod(fields.Method):
    """This class represents a field method with an alternative field to use as
    source for OpenAPI specification.

    :param spec_field:
        alternative field to use as source for openAPI specification
    :param serialize: serialization method used
    :param deserialize: deserialization method used
    """

    def __init__(
        self,
        spec_field: fields.Field,
        serialize: str | None = None,
        deserialize: str | None = None,
        **kwargs,
    ):
        self.field = spec_field
        super().__init__(
            serialize=serialize, deserialize=deserialize, **kwargs
        )


class FieldLookupNested(fields.Nested):
    """Nested field which will try to load an existing object.

    https://github.com/marshmallow-code/marshmallow-sqlalchemy/issues/117
    """

    default_error_messages = {
        "type": "Invalid input type. Expected a list.",
        "missing": "Could not find related object {input}.",
    }

    def get_instance_or_fail(self, data):
        instance = self.schema.get_instance(data)
        if not instance:
            self.fail("missing", input=data)
        return instance

    def _deserialize(self, value, attr, data):
        if self.many and not is_collection(value):
            self.fail("type", input=value, type=value.__class__.__name__)

        if self.many:
            return [self.get_instance_or_fail(item) for item in value]

        return self.get_instance_or_fail(value)


class BaseSchema(SQLAlchemyAutoSchema):
    """This is the base class for all schemas converting Json and SQL Models.

    There are multiple cases for schemas:

    * One schema for all CRU operations: then simply put all fields
    * Two schemas for update and retrieval:
      then make one schema with non-modifiable fields as `dump_only` in
      :attr:`Meta`
    * Two schemas for creation and retrieval:
      then make one schema with non-created fields as `dump_only` in
      :attr:`Meta` (can also add write-only fields in `load_only`)
    * Three schemas for creation, update and retrieval:
      split creation schema, then make other schema as an update/retrieval
      shared schema
    * Careful when associating a resource within another, as the update of one
      could update the other. If a single schema is used to represent both
      a resource as itself and the same one nested, the `id` field will be
      broken (non-writable for the resource as itself, writable for the nesting
      resource).

    .. note::

        * Fields with foreign keys are not created by default
        * Relationship fields are not created by default
        * FIelds added manually are not required and non-nullable by default
    """

    class Meta:
        sqla_session = db.session
        strict = True
        load_instance = True

    def get_future_field(self, data, key):
        """Return field value (from instance or serialized data).

        :param data: serialized data that will update instance
        :param key: field name
        :return: value
        """
        return data.get(
            key, None if self.instance is None else getattr(self.instance, key)
        )


class FieldPluck(fields.Pluck):
    """This class creates a pluck field for a SQL model.

    It creates a new schema for this model, only containing the field to pluck.

    :param model: SQL model class to pluck
    :param field_name: field to pluck
    """

    def __init__(self, model, field_name, **kwargs):
        meta_dict = {
            "model": model,
            "fields": (field_name,),
        }
        sch_dict = {"Meta": type("Meta", (BaseSchema.Meta,), meta_dict)}
        schema = type(f"Pluck{model.__name__}Schema", (BaseSchema,), sch_dict)
        super().__init__(schema, field_name, **kwargs)


class OneOfSchema(_OneOfSchema):
    """This class is a wrapper for
    :class:`marshmallow_oneofschema.OneOfSchema`.

    It adds one feature necessary to this backend:

    * Pass schema constructor args and kwargs to typed schemas
    """

    def __init__(self, *args, many=False, **kwargs):
        self._schema_args = args
        self._schema_kwargs = kwargs

        # Wrap schema initialization to add arguments
        self.type_schemas = {
            k: self.with_args(v) for k, v in self.type_schemas.items()
        }

        super(OneOfSchema, self).__init__(many=many)

    def with_args(self, schema):
        """Create a child schema constructor.

        Pass *args and **kwargs which were given when building the oneof schema
        instance.

        :param schema: child schema to prepare
        :return: child schema constructor
        """

        def init_schema(openapi=False):
            """Initialize a child schema.

            Pass *args and **kwargs which were given at parent schema
            instantiation.

            :param openapi:
                indicates whether this schema is created for populating
                an openAPI specification or not, defaults to ``False``
            """
            return schema(*self._schema_args, **self._schema_kwargs)

        return init_schema


class OneOfSchemaWithType(OneOfSchema):
    """This class is a wrapper for
    :class:`marshmallow_oneofschema.OneOfSchema`.

    It is to be used for schemas needing the type to appear in the openapi
    specification while being removed upon deserialization.
    This is typically the case for oneOf creation schemas.
    """

    def with_args(self, schema):
        """Create a child schema constructor.

        Pass *args and **kwargs which were given when building the oneof schema
        instance.

        :param schema: child schema to prepare
        :return: child schema constructor
        """

        def init_schema(openapi=False):
            """Initialize a child schema.

            Pass *args and **kwargs which were given at parent schema
            instantiation.
            Make parent `type_field` field writable by removing it from the
            `dump_only` fields and setting :attr:`dump_only` to ``False`` if
            `openapi` is ``True``.

            :param openapi:
                indicates whether this schema is created for populating
                an openAPI specification or not, defaults to ``False``
            """
            sch = schema(*self._schema_args, **self._schema_kwargs)
            if openapi:
                sch.fields.get(self.type_field).dump_only = False
                sch.dump_only = tuple(set(sch.dump_only) - {self.type_field})
            return sch

        return init_schema



from datetime import datetime
from typing import Self

from . import db


class BaseModelMixin:
    def save(self):
        """Commit the object to the database"""
        db.session.commit()

    def create(self):
        """Persist the current object in the database"""
        db.session.add(self)
        self.save()
        return self

    def delete(self):
        """Delete the current object from the database"""
        db.session.delete(self)
        db.session.commit()

    def on_modification(self):
        """Callback called when updating the data."""
        pass

    def update(self):
        """Update the current object"""
        self.on_modification()
        self.save()


class ModelMixin(BaseModelMixin):
    """CRUD mixin for the models"""

    id = db.Column(db.Integer, primary_key=True)

    def delete(self):
        super().delete()
        self.id = None

    @classmethod
    def get_by_id(cls, id) -> Self:
        """Returns the instance with the given id"""
        return cls.query.filter_by(id=id).first()


class TimestampMixin(ModelMixin):
    """CRUD mixin, extend with created_at and modified_at fields"""

    created_at = db.Column(db.DateTime)
    modified_at = db.Column(db.DateTime)

    def create(self):
        """Persist the object in the database, with a creation and modification
        date"""
        self.created_at = datetime.utcnow()
        self.modified_at = self.created_at
        return super().create()

    def on_modification(self):
        """Set the modification time"""
        super().on_modification()
        self.update_modified()

    def update_created(self):
        self.created_at = datetime.utcnow()

    def update_modified(self):
        self.modified_at = datetime.utcnow()
